# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from functools import cache
from itertools import chain
from typing import Any, Callable, Iterable, List, Optional

from .client import JsonApiClient
from .converter import (
    DataObjectConverter,
    JsonApiConverter
)
from .filter import ApiFilter
from .validate import validate, validate_id
from ..core import DataObject, DataSource, DataSourceFilter
from ..core.operator import (
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    Enricher,
    GroupStatter,
    Inserter,
    ListGetter,
    OperatorDict,
    PageGetter,
    RelationWriteMode,
    Relational,
    ReturnMode,
    Statter,
    Upserter,
)
from ..core.relationship import RelationshipConfig
from ..core.requested_fields import ReqFieldsTree, requested_fields_to_tree

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession


ClientFactory = Callable[[], JsonApiClient]
JsonConverterFactory = Callable[[], JsonApiConverter]
DOConverterFactory = Callable[[], DataObjectConverter]
FilterFactory = Callable[[], ApiFilter]


class ApiDataSource(
    DataSource,

    # the supported operators
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    Enricher,
    GroupStatter,
    Inserter,
    PageGetter,
    ListGetter,
    Relational,
    Statter,
    Upserter
):
    """
    A `DataSource` that connects to a remote API based upon
    `api_base2`.

    Developers should likely use `create_api_datasource`
    instead of this directly.
    """

    def __init__(
        self,
        client_factory: ClientFactory,
        json_converter_factory: JsonConverterFactory,
        do_converter_factory: DOConverterFactory,
        filter_factory: FilterFactory
    ) -> None:

        self.__client_factory = client_factory
        self.__jc_factory = json_converter_factory
        self.__dc_factory = do_converter_factory
        self.__filter_factory = filter_factory
        super().__init__({})
        self.write_batch_size = 100

    @property
    @cache
    def attribute_types(self) -> dict[str, dict[str, str]]:
        client = self.__client_factory()
        return client.config_attribute_types()

    @property
    @cache
    def attribute_metadata(self) -> dict[str, dict[str, dict[str, str | bool]]]:
        client = self.__client_factory()
        return client.config_attribute_metadata()

    @property
    @cache
    def supported_types(self) -> list[str]:
        return list(
            self.attribute_types.keys()
        )

    @property
    @cache
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        transfer = self.__client_factory().config_relationships()
        return self.__jc_factory().convert_relationship_config(
            transfer
        )

    @property
    @cache
    def write_mode(self) -> dict[str, RelationWriteMode]:
        transfer = self.__client_factory().config_write_mode()

        return {
            k: RelationWriteMode(v)
            for k, v in transfer.items()
        }

    @property
    @cache
    def return_mode(self) -> dict[str, ReturnMode]:
        transfer = self.__client_factory().config_return_mode()

        return {
            k: ReturnMode(v)
            for k, v in transfer.items()
        }

    @requested_fields_to_tree
    @validate('detailGet')
    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[OperableSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[Optional[DataObject]]:

        client = self.__client_factory()

        json_responses = (
            client.get_detail(
                object_type,
                id_,
                requested_tree=requested_tree,
            )
            for id_ in object_ids
        )
        json_converter = self.__jc_factory(object_type, requested_tree)
        return (
            json_converter.convert(r)
            if r is not None else None
            for r in json_responses
        )

    @requested_fields_to_tree
    @validate('listGet')
    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[OperableSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> tuple[Iterable[DataObject], int]:

        filter_string = self.__get_filter_string(object_filters)
        transfer = self.__client_factory().get_list_page(
            object_type,
            page_number,
            page_size,
            filter_string=filter_string,
            sort_string=sort_by,
            requested_tree=requested_tree,
        )

        return self.__jc_factory(object_type, requested_tree).convert_list(transfer)

    @requested_fields_to_tree
    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        if self.__can_cursor(object_type, object_filters):
            return self._get_list_by_cursor(
                object_type,
                object_filters,
                requested_tree=requested_tree,
            )
        else:
            return self.__get_list_regular(
                object_type,
                object_filters,
                requested_tree=requested_tree,
            )

    @validate('count')
    def get_count(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
    ) -> int:
        filter_string = self.__get_filter_string(object_filters)
        transfer = self.__client_factory().get_count(
            object_type,
            filter_string=filter_string
        )
        return self.__jc_factory().convert_count(transfer)

    @validate('stats')
    def get_stats(
        self,
        object_type: str,
        stats: Optional[List[str]] = [],
        stats_fields: Optional[List[str]] = [],
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None
    ) -> tuple[Iterable[DataObject], int]:

        filter_string = self.__get_filter_string(object_filters)
        transfer = self.__client_factory().get_stats(
            object_type,
            stats_string=','.join(stats),
            stats_fields_string=','.join(stats_fields),
            filter_string=filter_string
        )
        return self.__jc_factory().convert_stats(transfer)

    @validate('groupStats')
    def get_group_stats(
        self,
        object_type: str,
        group_by: List[str],
        stats_fields: List[str] = [],
        stats: List[str] = ['min', 'max'],
        object_filters: DataSourceFilter | None = None,
        session: OperableSession | None = None
    ) -> Iterable[OperatorDict[Any, int]]:

        filter_string = self.__get_filter_string(object_filters)
        transfer = self.__client_factory().get_group_stats(
            object_type,
            ','.join(group_by),
            stats_string=','.join(stats),
            stats_fields_string=','.join(stats_fields),
            filter_string=filter_string
        )
        return self.__jc_factory().convert_group_stats(transfer)

    @validate('cursor')
    def get_cursor_page(
        self,
        object_type: str,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        search_after: list[str] | None = None,
        session: Optional[OperableSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> tuple[Iterable[DataObject], list[str] | None]:

        filter_string = self.__get_filter_string(object_filters)
        transfer = self.__client_factory().get_cursor_page(
            object_type,
            page_size,
            search_after,
            filter_string=filter_string,
            requested_tree=requested_tree,
        )

        return self.__jc_factory(object_type, requested_tree).convert_cursor_page(transfer)

    @validate('delete')
    def delete(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[OperableSession] = None
    ) -> None:

        client = self.__client_factory()
        for object_id in object_ids:
            client.delete(object_type, object_id)

    @validate('upsert')
    def upsert_batch(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None,
        merge_collections: bool | None = None,
        **kwargs,
    ) -> Iterable[DataObject] | None:
        transfer = self.__dc_factory().convert_list(list(objects))
        returned = self.__client_factory().upsert(
            object_type,
            transfer,
            merge_collections=merge_collections,
        )
        if self.return_mode[object_type] == ReturnMode.POPULATED:
            converted, _ = self.__jc_factory(object_type).convert_list(returned)
            return converted
        return []  # when the underlying DataSource doesn't return anything

    @validate('relational', direct_object=True)
    @validate_id
    def get_recursive_relation(
        self,
        source: DataObject,
        relationship_hops: list[str],
        session: Optional[OperableSession] = None
    ) -> Optional[DataObject]:

        self.validate_to_one_recurse(source.type, relationship_hops)
        transfer = self.__client_factory().get_to_one_relation_recursive(
            source.type,
            source.id,
            relationship_hops
        )
        if transfer is None:
            return None
        return self.__jc_factory(source.type).convert(transfer)

    @validate('relational', direct_object=True)
    @validate_id
    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None
    ) -> Optional[DataObject]:

        return self.get_recursive_relation(
            source,
            [relationship_name]
        )

    @validate('relational', direct_object=True)
    @validate_id
    def get_to_many_relations_page(
        self,
        source: DataObject,
        relationship_name: str,
        page: int,
        page_size: int,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:

        transfer = self.__client_factory().get_to_many_relations_page(
            source.type,
            source.id,
            relationship_name,
            page,
            page_size
        )
        return self.__jc_factory(source.type).convert_list(transfer)

    @validate('relational', direct_object=True)
    @validate_id
    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject]:

        page_number = 1
        page_size = self.get_page_size()

        while True:
            page, _ = self.get_to_many_relations_page(
                source,
                relationship_name,
                page_number,
                page_size
            )

            next_page = list(page)
            if not next_page:
                return

            yield from next_page
            if len(next_page) < page_size:
                return
            page_number += 1

    @validate('insert')
    def insert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[OperableSession] = None
    ) -> Iterable[DataObject] | None:

        transfer = self.__dc_factory().convert_list(
            objects
        )
        returned = self.__client_factory().insert(
            object_type,
            transfer
        )
        if self.return_mode[object_type] == ReturnMode.POPULATED:
            converted, _ = self.__jc_factory(object_type).convert_list(returned)
            return converted

    @property
    @cache
    def supported_operations(self) -> dict[str, list[str]]:
        """
        The list of `Operator` ABC's implemented for each
        `object_type`.
        """

        client = self.__client_factory()
        transfer = client.config_operations()
        return self.__parse_operations(transfer)

    def __can_cursor(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None,
    ) -> bool:

        if 'cursor' not in self.supported_operations[object_type]:
            return False

        return self.can_use_cursor(object_type, object_filters)

    def __get_list_regular(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter],
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:

        page = 1
        page_size = self.get_page_size()
        client = self.__client_factory()

        jc_converter = self.__jc_factory(object_type, requested_tree)
        filter_string = self.__get_filter_string(object_filters)

        while True:
            transfer = client.get_list_page(
                object_type,
                page,
                page_size,
                filter_string=filter_string,
                requested_tree=requested_tree,
            )
            (results_page, _) = jc_converter.convert_list(transfer)

            yield from results_page
            if len(results_page) < page_size:
                break

            page += 1

    def __get_filter_string(
        self,
        object_filters: Optional[DataSourceFilter]
    ) -> Optional[str]:

        if object_filters is None:
            return None
        return self.__filter_factory().dumps(object_filters)

    def __parse_operations(
        self,
        transfer: dict[str, OperatorDict]
    ) -> dict[str, list[str]]:

        return {
            t: self.__join_operations(o)
            for t, o in transfer.items()
        }

    def __join_operations(
        self,
        operator_dict: OperatorDict
    ) -> list[str]:

        operators = chain(*list(operator_dict.values()))
        return list(operators)

    # We can't actually enrich as Updater is not implemented on this DataSource
    def enrich(
            self,
            source_object_type: str,
            source_objects: Iterable[DataObject],
            target_object_type: str):
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class StsBankedSampleToElasticSampleConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        ret = self._data_object_factory(
            'sample',
            data_object.sample.id,
            attributes={
                **data_object.attributes
            }
        )
        if data_object.category is not None:
            ret.banked_sample_category = data_object.category.name
        yield ret



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterable, Iterator, MutableMapping
from functools import reduce
from itertools import chain
from typing import Any, Dict, Optional, Tuple, Type

from sqlalchemy import BinaryExpression, cast, inspect, not_
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import MappedColumn, Query, Session, aliased
from sqlalchemy.orm.util import AliasedClass

from .model import Model
from ..core import DataSourceFilter


class AliasTrie(MutableMapping[str, 'AliasTrie']):

    def __init__(self, alias: AliasedClass[Model]) -> None:
        self.__alias = alias
        self.__dict: dict[str, AliasTrie] = {}

    @property
    def alias(self) -> AliasedClass[Model]:
        return self.__alias

    def __getitem__(self, key: str) -> AliasTrie:
        return self.__dict[key]

    def __setitem__(self, key: str, value: AliasTrie) -> None:
        self.__dict[key] = value

    def __delitem__(self, key: str) -> None:
        raise NotImplementedError()

    def __iter__(self) -> Iterator[str]:
        return iter(self.__dict)

    def __len__(self) -> int:
        return len(self.__dict)


class DatabaseFilter(ABC):
    """Filters an `sqlalchemy.orm` `Query` object"""

    @abstractmethod
    def filter(  # noqa A003
        self,
        query: Query[Model],
        tablename: str,
        model_dict: Dict[str, Type[Model]]
    ) -> Query[Model]:
        """Filter the Query object using the given model"""

    @abstractmethod
    def get_column(self, key: str) -> MappedColumn:
        """Gets the column for the given `DataObject` key"""

    @abstractmethod
    def add_field(self, field: str) -> None:
        """Adds a relation field to the filter, for joining later"""

    @abstractmethod
    def get_query(
        self, session: Session, base_model: type[Model]
    ) -> tuple[Query[Model], AliasedClass[Model]]:
        """Gets an aliased query"""


class DefaultDatabaseFilter(DatabaseFilter):
    """A reasonable-default database filter"""

    # TODO:
    # - sensible error checking/messages (e.g. if column does not exist)

    def __init__(
        self,
        datasource_filter: Optional[DataSourceFilter]
    ) -> None:

        self.__filter = datasource_filter
        self.__rel_keys: set[str] = set()

    def filter(  # noqa A003
        self,
        query: Query[Model],
        __tablename: str,
        __model_dict: Dict[str, Type[Model]]
    ) -> Query[Model]:

        self.__rel_keys.update(
            self.__generate_relational_keys()
        )

        self.__alias_trie = self.__build_alias_trie(
            self.__rel_keys,
        )

        query = self.__apply_joins(
            query,
            self.__alias_trie,
            self.__base_model,
        )

        if self.__filter is None:
            return query

        query = self.__filter_top_and_(query)
        query = self.__filter_top_exact(query)
        query = self.__filter_top_contains(query)
        query = self.__filter_top_in_list(query)
        query = self.__filter_top_range(query)

        return query

    def get_query(self, session: Session, base_model: Model) -> [Query[Model]]:
        self.__base_model = base_model

        return session.query(base_model)

    def __apply_joins(
        self,
        query: Query[Model],
        parent_trie: AliasTrie,
        parent_alias: AliasedClass[Model],
    ) -> Query[Model]:

        for part, trie in parent_trie.items():
            alias = trie.alias

            # Probably want an outerjoin() here
            query = query.join(alias, getattr(parent_alias, part))

            query = self.__apply_joins(
                query,
                trie,
                alias,
            )

        return query

    def __create_alias(
        self,
        parent_alias: AliasedClass[Model],
        relationship_name: str,
    ) -> AliasedClass[Model]:

        mapper = inspect(parent_alias).mapper
        rel_prop = mapper.relationships[relationship_name]
        target_model = rel_prop.mapper.class_
        return aliased(target_model)

    def add_field(self, field: str) -> None:
        self.__rel_keys.add(field)

    def __build_alias_trie(self, paths: Iterable[str]) -> AliasTrie:
        trie = AliasTrie(self.__base_model)

        for path in paths:
            parts = path.split('.')
            current_alias = self.__base_model
            current = trie
            for part in parts[:-1]:
                if part not in current:
                    step = AliasTrie(
                        self.__create_alias(
                            current_alias,
                            part,
                        )
                    )
                    current[part] = step
                    current = step
                else:
                    current = current[part]
                current_alias = current.alias

        return trie

    def __generate_relational_keys(self) -> Iterator[str]:
        if not self.__filter:
            return []

        chained = chain(
            self.__none_coalesce(self.__filter.exact),
            self.__none_coalesce(self.__filter.contains),
            self.__none_coalesce(self.__filter.in_list),
            self.__none_coalesce(self.__filter.range),
            self.__none_coalesce(self.__filter.and_),
            self.__filter_pointer_targets
        )
        return (
            k for k in chained
            if '.' in k
        )

    @property
    def __filter_pointer_targets(self) -> Iterator[str]:
        not_none = self.__none_coalesce(self.__filter.and_)
        for column_body in not_none.values():
            for term in column_body.values():
                if 'field' in term:
                    yield term['field']

    def __none_coalesce(self, in_: Optional[dict]) -> dict:
        return in_ if in_ is not None else {}

    def __filter_top_and_(self, query: Query[Model]) -> Query[Model]:
        if not self.__filter.and_:
            return query

        return reduce(
            lambda q, kv: self.__switch_and_term_dict(
                q,
                *kv
            ),
            self.__filter.and_.items(),
            query
        )

    def __switch_and_term_dict(
        self,
        query: Query[Model],
        column_key: str,
        term_dict: dict[str, dict[str, Any]]
    ) -> Query[Model]:

        return reduce(
            lambda q, kv: self.__switch_and_term(
                q,
                self.get_column(column_key),
                *kv
            ),
            term_dict.items(),
            query
        )

    def __switch_and_term(
        self,
        query: Query[Model],
        column: MappedColumn,
        op: str,
        term: dict[str, dict[str, Any]]
    ) -> Query[Model]:

        filter_dict = defaultdict(
            lambda: lambda *_: query,
            eq=self.__filter_eq,
            contains=self.__filter_contains,
            in_list=self.__filter_in_list,
            gt=self.__filter_gt,
            gte=self.__filter_gte,
            lt=self.__filter_lt,
            lte=self.__filter_lte,
            exists=self.__filter_exists
        )

        return filter_dict[op](query, column, term)

    def __parse_value_negate(
        self,
        term: dict[str, Any]
    ) -> tuple[Any, bool]:

        negate = term.get('negate', False)

        if 'field' in term:
            field = term['field']
            column = self.get_column(field)
            return column, negate
        else:
            return term.get('value'), negate

    def __filter_exists(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        _, negate = self.__parse_value_negate(term)

        if negate:
            return query.filter(
                column.is_(None)
            )
        else:
            return query.filter(
                column.is_not(None)
            )

    def __negatable_filter(
        self,
        query: Query[Model],
        expression: BinaryExpression,
        column: MappedColumn,
        negate: bool
    ) -> Query[Model]:

        if negate is True:
            return query.filter(
                (column.is_(None)) | not_(expression)
            )
        else:
            return query.filter(expression)

    def __filter_in_list(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column.in_(value)

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_contains(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)

        if self.__column_is_str(column):
            return self.__filter_contains_str(
                query,
                column,
                value,
                negate
            )
        else:
            return self.__filter_contains_list(
                query,
                column,
                value,
                negate
            )

    def __column_is_str(self, column: MappedColumn) -> bool:
        return column.type.python_type is str

    def __filter_contains_str(
        self,
        query: Query[Model],
        column: MappedColumn,
        value: str,
        negate: bool
    ) -> Query[Model]:

        ilike = self.__get_ilike_term(value)
        expression = column.ilike(ilike)

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_contains_list(
        self,
        query: Query[Model],
        column: MappedColumn,
        value: Any,
        negate: bool
    ) -> Query[Model]:

        jsonb_column = cast(column, JSONB)
        expression = jsonb_column.op('@>')([value])

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_eq(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column == value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_lt(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column < value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_lte(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column <= value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_gt(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column > value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_gte(
        self,
        query: Query[Model],
        column: MappedColumn,
        term: dict[str, Any]
    ) -> Query[Model]:

        value, negate = self.__parse_value_negate(term)
        expression = column >= value

        return self.__negatable_filter(
            query,
            expression,
            column,
            negate
        )

    def __filter_top_exact(self, query: Query[Model]) -> Query[Model]:
        exact_filters = self.__filter.exact
        if exact_filters is None:
            return query
        for k, v in exact_filters.items():
            exact_column = self.get_column(k)
            query = query.filter(exact_column == v)
        return query

    def __filter_top_contains(self, query: Query[Model]) -> Query[Model]:
        contains_filters = self.__filter.contains
        if contains_filters is None:
            return query
        for k, v in contains_filters.items():
            contains_column = self.get_column(k)
            term = self.__get_ilike_term(v)
            query = query.filter(contains_column.ilike(term))
        return query

    def __filter_top_in_list(self, query: Query[Model]) -> Query[Model]:
        in_filters = self.__filter.in_list
        if in_filters is None:
            return query
        for k, v in in_filters.items():
            in_column = self.get_column(k)
            query = query.filter(in_column.in_(v))
        return query

    def __filter_top_range(self, query: Query[Model]) -> Query[Model]:
        range_filters = self.__filter.range
        if range_filters is None:
            return query
        for k, v in range_filters.items():
            range_column = self.get_column(k)
            from_, to_ = self.__get_between_term(v)
            query = query.filter(range_column.between(from_, to_))
        return query

    def get_column(self, key: str, model: type[Any] | None = None) -> MappedColumn:
        model = self.__alias_trie.alias if model is None else model
        if key == 'id':
            return self.__get_id_column(model)
        elif '.' in key:
            return self.__get_relation_column(key)
        else:
            return self.__get_column_attr(model, key)

    def __get_id_column(self, model: type[Model]) -> MappedColumn:
        og_model: type[Model] = inspect(model).mapper.class_
        id_key = og_model.get_id_column_name()
        return self.__get_column_attr(model, id_key)

    def __get_column_attr(self, model: AliasedClass[Model], key: str) -> MappedColumn:
        for col in inspect(model).selectable.c:
            if col.key == key:
                return col
        msg = f"Failed to find column '{key}' in '{model}'"
        raise ValueError(msg)

    def __get_relation_column(
        self,
        key: str
    ) -> MappedColumn:

        (*initial, column) = key.split('.')

        if not initial:
            return self.get_column(key)

        trie = self.__alias_trie
        for i in initial:
            trie = trie[i]

        return self.get_column(
            column,
            model=trie.alias,
        )

    def __get_ilike_term(self, value: str) -> str:
        escaped = self.__escape_ilike(value)
        return f'%{escaped}%'

    def __get_between_term(self, value: Dict[str, Any]) -> Tuple[Any, Any]:
        from_ = value['from']
        to_ = value['to']
        return from_, to_

    def __escape_ilike(self, value: str) -> str:
        return (
            value.replace('\\', '\\\\').replace('%', '\\%').replace('_', '\\_')
        )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from tol.core import (
    core_data_object
)
from tol.google_sheets import (
    GoogleSheetDataSource
)


def bioscan_qc(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1_2lZ5nah_xadTQPhiTkNC5xNEa9VqK-u59Lu2okMD3U',
        'mappings': {
            'specimen': {
                'worksheet_name': 'SANGER_QC',
                'columns': {
                    'id': {
                        'heading': 'SPECIMEN_ID',
                        'type': 'str'
                    },
                    'sanger_qc_result': {
                        'heading': 'Sanger QC Result',
                        'type': 'str'
                    },
                    'sanger_qc_description': {
                        'heading': 'Sanger QC Description',
                        'type': 'str'
                    },
                },
                'header_row': 1,
                'data_start_row': 2
            },
            'uksi_entry': {
                'worksheet_name': 'UKSI_LIST',
                'columns': {
                    'id': {
                        'heading': 'RECOMMENDED_SCIENTIFIC_NAME',
                        'type': 'str'
                    },
                    'uksi_name_status': {
                        'heading': 'UKSI Name Status',
                        'type': 'str'
                    },
                },
                'header_row': 1,
                'data_start_row': 2
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Optional

from .filter_utils import FilterUtils


class StatsParameters:
    """
    Parses the parameters from a query string for a List GET
    endpoint.
    """

    def __init__(self, request_args: Dict[str, str]) -> None:
        self.__request_args = request_args

    @property
    def stats(self) -> Optional[int]:
        """The optional stats to return"""
        stats = self.__request_args.get('stats')
        if stats is None:
            return None

        return self._parse_to_list('stats', stats)

    @property
    def stats_fields(self) -> Optional[int]:
        """The optional stats to return"""
        stats_fields = self.__request_args.get('stats_fields')
        if stats_fields is None:
            return None

        return self._parse_to_list('stats_fields', stats_fields)

    @property
    def filter(self) -> Optional[str]:  # noqa A003
        """
        The optional filter JSON string.
        """
        filter_string = self.__request_args.get('filter')
        if filter_string is None:
            return None

        return FilterUtils.parse_to_datasource_filter('filter', filter_string)

    @property
    def _args(self) -> dict[str, Any]:
        return self.__request_args

    def _parse_to_list(self, __key: str, __value: str) -> int:
        return __value.split(',') if __value != '' else []


class GroupStatsParameters(StatsParameters):

    @property
    def group_by(self) -> list[str]:
        group_bys = self._args.get('group_by')
        if group_bys is None:
            return None

        return self._parse_to_list('group_by', group_bys)



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from tol.core import (
    core_data_object
)
from tol.google_sheets import (
    GoogleSheetDataSource
)


def bioscan_extra(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1XiKKnz8O-GcQ5ww19m1_1Gk7shidUTIvzidY47-F3hs',
        'mappings': {
            'pantheon_species': {
                'worksheet_name': 'NEW_PANTHEON',
                'columns': {
                    'id': {
                        'heading': 'Species',
                        'type': 'str'
                    },
                    'vernacular': {
                        'heading': 'Pantheon:Vernacular',
                        'type': 'str'
                    },
                    'current_conservation_status': {
                        'heading': 'Pantheon:Conservation Status Description',
                        'type': 'str'
                    },
                    'larval_feeding_guild': {
                        'heading': 'Pantheon:Larval feeding guild',
                        'type': 'str'
                    },
                    'adult_feeding_guild': {
                        'heading': 'Pantheon:Adult feeding guild',
                        'type': 'str'
                    },
                    'broad_biotope_habitat_resources': {
                        'heading': 'Pantheon:Broad Biotope, Habitat, Resources',
                        'type': 'str'
                    },
                    'specific_assemblage_type': {
                        'heading': 'Pantheon:Specific assemblage type',
                        'type': 'str'
                    },
                    'link_to_assemblage': {
                        'heading': 'Link to assemblage',
                        'type': 'str'
                    },
                    'associations': {
                        'heading': 'Pantheon:Associations',
                        'type': 'str'
                    },
                },
                'header_row': 1,
                'data_start_row': 2
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Any, Dict

from tol.core import DataObject, DataSource
from tol.core.validate import Validator
from tol.sources.ena import ena


class EnaSubmittableValidator(Validator):
    """
    Validates that a stream of `DataObject` instances
    contains field that is part of a list.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field_name: str

    __slots__ = ['__config', '__ena_datasource', '__cached_species']
    __config: Config
    __ena_datasource: DataSource
    __cached_species: Dict[str, Any]

    def __init__(
        self,
        config: Config,
        ena_datasource: DataSource | None = ena(),  # For testing
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = config
        self.__ena_datasource = ena_datasource
        self.__cached_species = {}

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        taxon_id = obj.get_field_by_name(self.__config.field_name)
        if taxon_id not in self.__cached_species:
            ena_taxon = self.__ena_datasource.get_one('submittable_taxon', taxon_id)
            if ena_taxon:
                self.__cached_species[taxon_id] = ena_taxon
        if taxon_id not in self.__cached_species:
            self.add_error(
                object_id=obj.id,
                detail=f'Field {self.__config.field_name} value '
                       f'"{taxon_id}" not found in ENA',
                field=self.__config.field_name,
            )
            return

        if (
            str(taxon_id) == '32644'
            and self.__cached_species[taxon_id].rank == 'species'
            and self.__cached_species[taxon_id].submittable
        ):
            return
        elif not self.__cached_species[taxon_id].binomial:
            self.add_error(
                object_id=obj.id,
                detail=f'Field {self.__config.field_name} value '
                       f'"{taxon_id}" taxon species name is not-binomial',
                field=self.__config.field_name,
            )
        elif not self.__cached_species[taxon_id].submittable:
            self.add_error(
                object_id=obj.id,
                detail=f'Field {self.__config.field_name} value '
                       f'"{taxon_id}" is not submittable in ENA',
                field=self.__config.field_name,
            )
        elif self.__cached_species[taxon_id].rank == 'subspecies':
            self.add_warning(
                object_id=obj.id,
                detail='Please consider onboarding your sample at species rank '
                       '(provide taxon and species name); subspecies information '
                       'can be recorded in the infra_epithet field on the manifest.',
                field=self.__config.field_name,
            )
        elif self.__cached_species[taxon_id].rank != 'species':
            self.add_error(
                object_id=obj.id,
                detail=f'Field {self.__config.field_name} value '
                       f'"{taxon_id}" has rank '
                       f'"{ self.__cached_species[taxon_id].rank}" which is not acceptable',
                field=self.__config.field_name,
            )



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import re
from functools import cache
from typing import Dict, Iterable, Tuple

import pandas as pd

import requests

from ..core import (
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter
)
from ..core.operator import PageGetter


class TreevalDataSource(
    DataSource,
    PageGetter,
):

    def __init__(self, config: Dict):
        # uri, user, password
        super().__init__(config, expected=['url', 'api_token'])

    def _build_jira_query(self):
        return '{"jql":"status = curation and project in (RC,GRIT)","maxResults":1000,\
            "fields":["key", "priority", "fields", "updated", "customfield_12200",\
                 "customfield_11676", "customfield_11677", "summary", "assignee",\
                     "attachment", "description", "customfield_11643", "customfield_11605",\
                     "customfield_12800", "customfield_12802"]}'

    def _execute_jira_query(self, query):
        response = requests.post(
            url=f'https://{self.url}/rest/api/latest/search',
            headers={'Authorization': 'Bearer ' + self.api_token,
                     'Content-Type': 'application/json'},
            data=query
        )

        if (response.status_code != 200):
            raise DataSourceError(title='Cannot connect to JIRA',
                                  detail=f"(status code '{str(response.status_code)}')'")

        return response.json()

    def _parse_jira_output(self, response_text):

        issues = map(self._get_values_from_issue, response_text['issues'])

        return pd.DataFrame(issues)

    def _parse_decontamination_report(self, decon_data):
        tot_removed = ''
        tot_removed_pc = ''
        count_removed = ''
        count_removed_pc = ''
        largest_removed = ''
        is_abnormal = 'false'

        tot_removed_regex = \
            re.compile(r'Total length of scaffolds removed: ([0-9,]+) \(([0-9\.]+) %\)')
        count_removed_regex = re.compile(r'Scaffolds removed: ([0-9,]+) \(([0-9\.]+) %\)')
        largest_removed_regex = re.compile(r'Largest scaffold removed: \(([0-9,]+)\)')

        tot_rem = tot_removed_regex.search(decon_data)
        count_rem = count_removed_regex.search(decon_data)
        largest_rem = largest_removed_regex.search(decon_data)

        if tot_rem:
            tot_removed = tot_rem.group(1)
            tot_removed_pc = tot_rem.group(2)

        if count_rem:
            count_removed = count_rem.group(1)
            count_removed_pc = count_rem.group(2)

        if largest_rem:
            largest_removed = largest_rem.group(1)

        if 'Abnormal contamination report' in decon_data:
            is_abnormal = 'true'

        return tot_removed, tot_removed_pc, count_removed, \
            count_removed_pc, largest_removed, is_abnormal

    def _parse_description_for_stats(self, description):

        scaffold_l90 = '-'
        contig_l90 = '-'

        scaffold_l90_regex = re.compile(r'SCAFFOLD[ \t\n\r\f\v]N90 = [0-9]+, L90 = ([0-9]+)')
        contig_l90_regex = re.compile(r'CONTIG[ \t\n\r\f\v]N90 = [0-9]+, L90 = ([0-9]+)')

        sl90 = scaffold_l90_regex.search(description)
        cl90 = contig_l90_regex.search(description)

        if sl90:
            scaffold_l90 = sl90.groups(1)

        if cl90:
            contig_l90 = cl90.groups(1)

        return scaffold_l90, contig_l90

    def _get_values_from_issue(self, issue):

        key = issue['key']

        fields = issue['fields']
        priority = fields['priority']
        updated = pd.Timestamp(fields['updated'])
        species_name = self._parse_species_name(fields['customfield_11676'])

        tolid = self._parse_species_id(fields['summary'])
        tolid_assem = fields['customfield_11643']

        expected_karyotype = fields['customfield_11605']
        if not expected_karyotype:
            expected_karyotype = '-'

        con_filname = fields['customfield_11677']
        file_struct = con_filname.split('/')
        tolqc_project = file_struct[5]

        if tolqc_project == 'meier':
            tolqc_clade = file_struct[8]

        elif tolqc_project == 'badass':
            tolqc_project = 'lawniczak'
            tolqc_clade = 'badass'
        else:
            tolqc_clade = file_struct[7]

        species_name_parts = species_name.split(' ')
        if len(species_name_parts) > 1:
            tolqc_species_name = f'{species_name_parts[0]}_{species_name_parts[1]}'
        else:
            tolqc_species_name = ''

        # Treeval link
        treeval_val = fields['customfield_12800']

        if not treeval_val or 'jb_scaffold' not in treeval_val:
            treeval_val = \
                '{"jbrowse": "","jb_scaffold": "","start": "","btk_pr": "","btk_hp": "",\
                "higlass": "","taxon_id": ""}'

        treeval_data = json.loads(treeval_val)

        # Stats from description
        description = str(fields['description'])
        scaffold_l90, contig_l90 = self._parse_description_for_stats(description)

        # Parse decontamination
        decon_data = str(fields['customfield_12802'])
        tot_removed, tot_removed_pc, count_removed, count_removed_pc, \
            largest_removed, is_abnormal = self._parse_decontamination_report(decon_data)

        # Assignee
        assignee = fields['assignee']

        if assignee:
            display_name = assignee['displayName']
        else:
            display_name = 'Unassigned'

        # Hi-C Contact Map image
        hic_plot_path = ''
        if 'hic_plot' in treeval_data.keys():
            if treeval_data['hic_plot'] == 'Y':
                hic_plot_path = \
                    f'https://treeval.cog.sanger.ac.uk/pretextsnapshot_{tolid_assem}.png'

        # K-mer spectra image
        kmer_plot_path = ''
        if 'kmer_plot' in treeval_data.keys():
            if treeval_data['kmer_plot'] == 'Y':
                kmer_plot_path = \
                    f'https://treeval.cog.sanger.ac.uk/kmerspectra_{tolid_assem}.png'

        # jBrowse link
        if 'jbrowse' in treeval_data.keys():
            if treeval_data['jbrowse']:

                tolid = treeval_data['jbrowse']
                scaff = treeval_data['jb_scaffold']
                server = treeval_data['jb_server']

                if server == 'dev':
                    server_url = 'tol-dev'
                else:
                    server_url = 'tol'

                jbrowse_link = ('http://jbrowse.' + server_url
                                + '.sanger.ac.uk/jbrowse2/?config=config.json'
                                + '&assembly=' + tolid + '&session=spec-'
                                + '{%22views%22:[{%22assembly%22:%22'
                                + tolid + '%22,%22loc%22:%22' + scaff
                                + '%22,%22type%22:%22LinearGenomeView%22,%22tracks%22:[%22'
                                + tolid + '-ReferenceSequenceTrack%22]}]}')

            else:
                jbrowse_link = ''
        else:
            jbrowse_link = ''

        # BTK primary link
        if 'btk_pr' in treeval_data.keys():
            btk_pr = treeval_data['btk_pr']
            if btk_pr:
                btk_pr_link = (f'https://grit-btk.tol.sanger.ac.uk/view/{btk_pr}'
                               + f'/dataset/{btk_pr}.fa.ascc/blob')
            else:
                btk_pr_link = ''
        else:
            btk_pr_link = ''

        # BTK haplotigs link
        if 'btk_hp' in treeval_data.keys():
            btk_hp = treeval_data['btk_hp']
            if btk_hp:
                btk_hp_link = (f'https://grit-btk.tol.sanger.ac.uk/view/{btk_hp}'
                               + f'/dataset/{btk_hp}.fa.ascc/blob')
            else:
                btk_hp_link = ''
        else:
            btk_hp_link = ''

        # GoaT link
        if 'taxon_id' in treeval_data.keys():
            taxon_id = treeval_data['taxon_id']
            if taxon_id:
                goat_link = (f'https://goat.genomehubs.org/record?recordId={taxon_id}'
                             + '&result=taxon&taxonomy=ncbi')
            else:
                goat_link = ''
        else:
            goat_link = ''

        # HiGlass link
        if 'higlass' in treeval_data.keys():
            higlass_id = treeval_data['higlass']
            if higlass_id:
                higlass_link = f'https://grit-higlass.tol.sanger.ac.uk/l/?d={higlass_id}'
            else:
                higlass_link = ''
        else:
            higlass_link = ''

        # Start date
        if treeval_data['start'] != '':
            added_to_curation_date = pd.Timestamp(treeval_data['start'])
        else:
            added_to_curation_date = pd.Timestamp('1970-01-01T00:00:00.000+0100')

        # ToLQC link
        if tolqc_project not in ('tol-nematodes', 'genomeark'):
            tolqc_link = (f'https://tolqc.cog.sanger.ac.uk/{tolqc_project}/'
                          + f'{tolqc_clade}/{tolqc_species_name}/')
        else:
            tolqc_link = ''

        return {'tolid': tolid,
                'species_name': species_name,
                'priority': str(priority['id']),
                'jira_issue': key,
                'jira_issue_url': f'https://{self.url}/browse/{key}',
                'jira_issue_last_updated': updated,
                'added_to_curation': added_to_curation_date,
                'jbrowse_url': jbrowse_link,
                'assignee': display_name,
                'goat_url': goat_link,
                'higlass_url': higlass_link,
                'btk_pri_url': btk_pr_link,
                'btk_hap_url': btk_hp_link,
                'tolqc_url': tolqc_link,
                'hic_plot': hic_plot_path,
                'kmer_plot': kmer_plot_path,
                'scaffold_l90': scaffold_l90,
                'contig_l90': contig_l90,
                'expected_karyotype': str(expected_karyotype),
                'total_scaffolds_removed': str(tot_removed),
                'total_scaffolds_removed_pc': str(tot_removed_pc),
                'scaffolds_removed_count': str(count_removed),
                'scaffolds_removed_count_pc': str(count_removed_pc),
                'largest_scaffold_removed': largest_removed,
                'contamination_is_abnormal': is_abnormal
                }

    def _parse_species_name(self, species_name):
        if species_name:

            # Trim unused common name
            suffix = ' ()'
            if species_name.endswith(suffix):
                species_name = species_name[:-len(suffix)]

            return species_name
        else:
            return ''

    def _parse_species_id(self, summary):
        species_id = str(summary)

        if species_id != '':
            species_id = species_id.replace(' GenomeArk assembly', '')
            species_id = species_id.replace(' ERGA assembly', '')
            species_id = species_id.replace(' Darwin assembly', '')
            species_id = species_id.replace(' faculty assembly', '')
            species_id = species_id.replace(' ASG assembly', '')
            species_id = species_id.replace(' VGP assembly', '')
            species_id = species_id.replace(' external assembly', '')
            species_id = species_id.replace(' TOL assembly', '')
            species_id = species_id.replace(' assembly', '')

            return species_id
        else:
            return ''

    def _apply_contains_filter_to_specimens(self, object_filters, specimens):

        if 'tolid' in object_filters:
            specimens = specimens[specimens['tolid']
                                  .str.contains(object_filters['tolid'])]

        if 'species_name' in object_filters:
            specimens = specimens[specimens['species_name']
                                  .str.contains(object_filters['species_name'])]

        if 'jira_issue' in object_filters:
            specimens = specimens[specimens['jira_issue']
                                  .str.contains(object_filters['jira_issue'])]

        if 'jira_issue_link' in object_filters:
            specimens = specimens[specimens['jira_issue_link']
                                  .str.contains(object_filters['jira_issue_link'])]

        if 'jbrowse_link' in object_filters:
            specimens = specimens[specimens['jbrowse_link']
                                  .str.contains(object_filters['jbrowse_link'])]

        if 'assignee' in object_filters:
            specimens = specimens[specimens['assignee']
                                  .str.contains(object_filters['assignee'])]

        if 'jbrowse_status' in object_filters:
            specimens = specimens[specimens['jbrowse_status']
                                  .str.contains(object_filters['jbrowse_status'])]

        return specimens

    def _apply_range_filter_to_specimens(self, object_filters, specimens):

        if 'jira_issue_last_updated' in object_filters:
            last_updated_range = object_filters['jira_issue_last_updated']
            specimens = specimens[(specimens['jira_issue_last_updated']
                                  > pd.Timestamp(last_updated_range['from']))
                                  & (specimens['jira_issue_last_updated']
                                      < pd.Timestamp(last_updated_range['to']))]

        if 'added_to_curation' in object_filters:
            added_to_curation_range = object_filters['added_to_curation']
            specimens = specimens[(specimens['added_to_curation']
                                  > pd.Timestamp(added_to_curation_range['from']))
                                  & (specimens['added_to_curation']
                                      < pd.Timestamp(added_to_curation_range['to']))]

        return specimens

    def _apply_sort_to_specimens(self, sort_by, specimens):

        if sort_by is None:
            specimens = specimens.sort_values(by=['added_to_curation'], ascending=[False])
        else:
            if sort_by.startswith('-'):
                column_name = sort_by[1:]
                sort_direction = False
            else:
                column_name = sort_by
                sort_direction = True

            specimens = specimens.sort_values(by=[column_name, 'added_to_curation'],
                                              ascending=[sort_direction, False])

        return specimens

    def get_list_page(
        self,
        object_type: str,
        page: int,
        object_filters: DataSourceFilter = None,
        sort_by: str = None,
        page_size: int = None,
        **kwargs
    ) -> Tuple[Iterable[DataObject], int]:

        query = self._build_jira_query()
        response = self._execute_jira_query(query)

        # Convert raw jira output data to visible outputs.
        specimens = self._parse_jira_output(response)

        # object_filters needs to be dict - currently string
        object_filters_dict = json.loads(object_filters)

        # Filter
        if object_filters_dict and len(object_filters_dict.keys()) > 0:

            if 'contains' in object_filters_dict.keys():
                specimens = self._apply_contains_filter_to_specimens(
                    object_filters_dict['contains'], specimens)

            if 'range' in object_filters_dict.keys():
                specimens = self._apply_range_filter_to_specimens(
                    object_filters_dict['range'], specimens)

        # Sort
        specimens = self._apply_sort_to_specimens(sort_by, specimens)

        # specimens = specimens[specimens['jbrowse_url']
        #                         .str.contains('sanger')]

        full_len = len(specimens)

        if not page_size:
            page_size = 50

        if not page:
            page = 1

        end_val = int(page) * int(page_size)
        start_val = end_val - int(page_size)

        if len(specimens) < end_val:
            end_val = len(specimens)

        # Filter to current page
        specimens = specimens.iloc[start_val:end_val, ]

        return (specimens.to_dict('records'), full_len)

    def get_specimens_for_treeval(self, page_number, page_size, filter_, sort_by):

        specimens_page, total_specimen_count = self.get_list_page(
            object_type='specimen',
            page=page_number,
            object_filters=filter_,
            sort_by=sort_by,
            page_size=page_size
        )

        return {'total': total_specimen_count, 'data': specimens_page}

    def get_specimen_for_treeval(self, tolid):
        return self.get_specimens_for_treeval(1, 1, f'[tolid={tolid}]', 'tolid')[0]

    def get_by_id():
        raise NotImplementedError()

    def get_list():
        raise NotImplementedError()

    def get_aggregations():
        raise NotImplementedError()

    @property
    def attribute_types(self):
        raise NotImplementedError()

    @property
    @cache
    def supported_types(self):
        raise NotImplementedError()



# SPDX-FileCopyrightText: 2022 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from collections.abc import Callable, Iterable
from datetime import datetime
from typing import Any

from cachetools.func import ttl_cache

from caseconverter import (
    kebabcase,
    snakecase
)

from elasticsearch import (Elasticsearch, helpers)

from .client import ElasticClient
from .converter import (
    ElasticApiConverter,
    ElasticUpdateInputConverter,
    ElasticUpsertInputConverter,
)
from .filter import ElasticFilterConverter
from .parser import ElasticUpdateInputResource, ElasticUpsertInputResource
from ..core import (
    AttributeMetadata,
    DataId,
    DataObject,
    DataSource,
    DataSourceError,
    DataSourceFilter,
    DefaultAttributeMetadata,
    GroupStatterDataLoader
)
from ..core.operator import (
    Aggregator,
    Counter,
    Cursor,
    DetailGetter,
    Enricher,
    GroupStatter,
    ListGetter,
    PageGetter,
    RelationWriteMode,
    Relational,
    Statter,
    Summariser,
    Updater,
    Upserter,
)
from ..core.operator.updater import DataObjectUpdate
from ..core.relationship import (
    RelationshipConfig
)
from ..core.requested_fields import ReqFieldsTree, requested_fields_to_tree

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession

ElasticClientFactory = Callable[[], ElasticClient]
ElasticConverterFactory = Callable[[], ElasticApiConverter]
DataObjectConverterFactory = Callable[[], ElasticUpsertInputConverter]
DataObjectUpdateConverterFactory = Callable[[], ElasticUpdateInputConverter]


class ElasticDataSource(
    DataSource,
    Cursor,
    Summariser,
    DetailGetter,
    Enricher,
    PageGetter,
    ListGetter,
    Aggregator,
    Relational,
    Updater,
    Upserter,
    Counter,
    Statter,
    GroupStatter
):
    def __init__(self, config: dict,
                 client_factory: ElasticClientFactory,
                 elastic_converter_factory: ElasticConverterFactory,
                 data_object_converter_factory: DataObjectConverterFactory,
                 data_object_update_converter_factory: DataObjectUpdateConverterFactory,
                 attribute_metadata: AttributeMetadata = DefaultAttributeMetadata,
                 relationship_cfg: dict[str, RelationshipConfig] | None = None,
                 runtime_fields: dict[str, Any] = {},
                 **kwargs):
        del kwargs
        super().__init__(
            config,
            expected=['uri', 'user', 'password', 'index_prefix'],
            attribute_metadata=attribute_metadata,
        )
        self._client_factory = client_factory
        self._elastic_converter_factory = elastic_converter_factory
        self._data_object_converter_factory = data_object_converter_factory
        self._data_object_update_converter_factory = data_object_update_converter_factory
        """
        relationship_cfg is also supported if we want to handle relationships
        Only FKs pointing to IDs are currently supported
        """
        attribute_metadata.host = self
        self.runtime_fields = runtime_fields
        self._initialise_elasticsearch()
        self.__lazy = False
        self._relationship_cfg = relationship_cfg

    @property
    def lazy_fetch(self) -> bool:
        """
        If `True`, enriched fields will be consulted.

        If `False`, enriched fields will be ignored, and the relation object
        directly fetched by ID, every time.
        """

        return self.__lazy

    @lazy_fetch.setter
    def lazy_fetch(self, new_val: bool) -> None:
        self.__lazy = new_val

    @property
    def _default_write_mode(self) -> RelationWriteMode:
        return RelationWriteMode.FUSED

    def _initialise_elasticsearch(self):
        self.es = Elasticsearch(
            self.uri,
            http_auth=(self.user, self.password),
            timeout=30,
            max_retries=10,
            retry_on_timeout=True
        )
        self.helpers = helpers

    def get_cursor_page(
        self,
        object_type: str,
        page_size: int | None = None,
        object_filters: DataSourceFilter | None = None,
        search_after: list[str] | None = None,
        session: OperableSession | None = None,
        **kwargs,
    ) -> tuple[Iterable[DataObject], list[str] | None]:
        del session, kwargs

        resp = self.__get_page_response(
            object_type,
            object_filters,
            'id',
            page_size,
            search_after=search_after
        )

        return self.__format_cursor_response(resp)

    def upsert(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        chunk_size: int = 100,
        id_func=lambda x: x.id,
        provenance: str = '',
        merge_collections: bool | None = None,
        **kwargs
    ) -> None:
        del kwargs

        if merge_collections is False:
            msg = 'ElasticDataSource does not support turning off merge_collections'
            raise DataSourceError(msg)

        index = self.__get_index_or_alias(object_type)
        converter = self._data_object_converter_factory()
        (no_of_operations, no_of_errors) = \
            self.helpers.bulk(
                self.es,
                converter.convert(
                    ElasticUpsertInputResource(
                        index=index,
                        objects=objects,
                        id_func=id_func,
                        field_prefix=provenance
                    )
                ),
                stats_only=True,
                chunk_size=chunk_size
        )
        if no_of_errors > 0:
            raise DataSourceError(f'{no_of_errors} errors encountered '
                                  f'upserting {no_of_operations} objects')

    def update(
        self,
        object_type: str,
        updates: Iterable[DataObjectUpdate],
        provenance: str = '',
        candidate_key: Iterable[str] = [],
        **kwargs
    ) -> None:

        # This tries to find an object in the DataSource that matches
        # the candidate key. If found it will perform the update

        index = self.__get_index_or_alias(object_type)
        real_index_name = self._get_indices().get(index)
        converter = self._data_object_update_converter_factory()
        for (_, update) in updates:
            # converter takes update, returns candidate key and body

            # We can get the candidate key dynamically from the actual update
            if 'candidate_key_func' in kwargs:
                candidate_key = kwargs['candidate_key_func'](update)
            self.es.update_by_query(
                index=real_index_name,
                body=converter.convert(ElasticUpdateInputResource(object_type=object_type,
                                                                  update=update,
                                                                  field_prefix=provenance,
                                                                  candidate_key=candidate_key)),
                conflicts='proceed',
                wait_for_completion=False
            )

    def _summarise(
        self,
        summary: DataObject,
        ext_and: dict[str, Any] | None = None,
    ) -> None:

        loader = GroupStatterDataLoader(
            self,
            self,
            [],
            summary.source_object_type,
            summary.destination_object_type,
            'Unmanaged summariser (no audit)',
            object_filters=self._mix_in_ext_and(
                summary.object_filters,
                ext_and,
            ),
            group_statter_group_by=summary.group_by,
            group_statter_stats_fields=summary.stats_fields,
            group_statter_stats=summary.stats,
        )
        loader.load(provenance=summary.provenance_override)

    def __format_cursor_response(
        self,
        resp: dict[str, Any]
    ) -> tuple[Iterable[DataObject], list[str] | None]:

        hits = list(resp['hits']['hits'])
        if not hits:
            return [], None

        search_after = hits[-1]['sort']
        objs = self._elastic_converter_factory().convert_list(hits)

        return objs, search_after

    def __get_index_or_alias(self, object_type: str) -> str:
        return f'{self.index_prefix}-{kebabcase(object_type)}'

    def __get_object_type(self, index: str) -> str:
        start = len(self.index_prefix) + 1
        return snakecase(index[start:])

    def _field_or_keyword(self, object_type: str, name: str):
        """
        Map our field format to Elastic's format
        """
        if name == 'id':
            return 'uid.keyword'
        # Runtime fields don't behave the same as text fields
        if object_type in self.runtime_fields and name in self.runtime_fields[object_type]:
            return name
        # An attribute of the object
        if name in self.attribute_types[object_type]:
            field_type = self.attribute_types[object_type][name]
            if field_type == 'str':
                return f'{name}.keyword'
        if '.' in name:
            rc = self.relationship_config[object_type]
            relationship_name, attribute = name.split('.')[0], name.split('.')[1]
            if attribute == 'id':
                return f'{name}.keyword'
            relationship_object_type = rc.to_one[relationship_name]
            attribute_type = self.attribute_types[relationship_object_type][attribute]
            if attribute_type == 'str':
                return f'{name}.keyword'
        return name

    def _prepare_get_parameters(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None,
        sort_by: str | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> tuple[str | None, dict, list[Any] | None, dict | None]:
        """
        Prepares the real_index_name, query, fields, and runtime_mappings,
        needed for all get operations.
        `fields` and `runtime_mappings` are filtered by the fields in the requested tree,
        if one was provided.
        """

        # Prepare real_index_name
        index = self.__get_index_or_alias(object_type)
        real_index_name = self._get_indices().get(index)

        # Prepare query
        query = ElasticFilterConverter(self).convert(object_type, object_filters)

        # Prepare fields to request and their runtime_mappings.
        # Filter runtime fields to have only those in the requested tree.
        runtime_mappings = (
            dict(self.runtime_fields[object_type])
            if object_type in self.runtime_fields
            else None
        )
        fields = list(runtime_mappings.keys()) if runtime_mappings is not None else None
        if requested_tree is not None and fields is not None and runtime_mappings is not None:
            # Filter fields to fetch based on whether they're in the requested tree
            fields = list(filter(
                lambda field: (
                    requested_tree.has_attribute(field)
                    or (
                        object_filters is not None and object_filters.and_ is not None
                        and field in object_filters.and_.keys()
                    )
                    or sort_by is not None and field in sort_by
                ),
                fields,
            ))

            # Only allow the runtime mappings of these fields
            runtime_mappings = {key: runtime_mappings[key] for key in fields}

        return (
            real_index_name,
            query,
            fields,
            runtime_mappings,
        )

    @requested_fields_to_tree
    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[DataId],
        requested_tree: ReqFieldsTree | None = None,
        **kwargs
    ) -> Iterable[DataObject]:
        del kwargs
        f = DataSourceFilter()
        f.and_ = {'_id': {'in_list': {'value': object_ids}}}
        # get_by_id is expected to return objects in the order they were asked for
        # or None if not found, hence the following rearrangement.
        return self.sort_by_id(
            self.get_list(object_type, object_filters=f, requested_tree=requested_tree),
            object_ids
        )

    @requested_fields_to_tree
    def get_list_page(
        self,
        object_type: str,
        page: int,
        object_filters: DataSourceFilter | None = None,
        sort_by: str | None = None,
        page_size: int | None = None,
        requested_tree: ReqFieldsTree | None = None,
        **kwargs
    ) -> tuple[Iterable[DataObject], int]:
        del kwargs

        resp = self.__get_page_response(
            object_type,
            object_filters,
            sort_by,
            page_size,
            page=page,
            requested_tree=requested_tree,
        )

        return (
            self._elastic_converter_factory().convert_list(
                resp['hits']['hits']
            ),
            resp['hits']['total']['value']
        )

    def __get_page_response(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None,
        sort_by: str | None,
        page_size: int | None,
        page: int | None = None,
        search_after: list[Any] | None = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> dict[str, Any]:
        real_index_name, query, fields, runtime_mappings = self._prepare_get_parameters(
            object_type=object_type,
            object_filters=object_filters,
            sort_by=sort_by,
            requested_tree=requested_tree,
        )
        sort = self._build_elasticsearch_sort(object_type, sort_by)
        if page_size is None:
            page_size = self.get_page_size()
        from_ = (page - 1) * page_size if page is not None else None
        return self.es.search(
            from_=from_,
            size=page_size,
            index=real_index_name,
            query=query,
            sort=sort,
            fields=fields,
            runtime_mappings=runtime_mappings,
            search_after=search_after
        )

    def _build_elasticsearch_sort(
        self,
        object_type: str,
        sort_by: str | None
    ) -> list[dict[str, str]]:
        default_sort = {'uid.keyword': 'asc'}
        if sort_by is None:
            return [default_sort]
        if sort_by == '-id':
            return self.__build_uid_sort(True)
        if sort_by == 'id':
            return self.__build_uid_sort(False)
        return [
            self.__build_sort(object_type, sort_by),
            default_sort
        ]

    def __build_uid_sort(
        self,
        desc: bool
    ) -> list[dict[str, str]]:

        order = 'desc' if desc else 'asc'
        return [
            {
                'uid.keyword': {
                    'order': order,
                    'unmapped_type': 'keyword'
                }
            }
        ]

    def __build_sort(
        self,
        object_type: str,
        sort_by: str
    ) -> dict[str, str]:

        if sort_by.startswith('-'):
            field = self._field_or_keyword(object_type, sort_by[1:])
            order = 'desc'
        else:
            field = self._field_or_keyword(object_type, sort_by)
            order = 'asc'

        return {field: order}

    @requested_fields_to_tree
    def get_list(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None = None,
        session: OperableSession | None = None,
        requested_tree: ReqFieldsTree | None = None,
        **kwargs
    ) -> Iterable[DataObject]:
        del session, kwargs

        real_index_name, query, fields, runtime_mappings = self._prepare_get_parameters(
            object_type=object_type,
            object_filters=object_filters,
            requested_tree=requested_tree,
        )
        generator = self.helpers.scan(self.es,
                                      index=real_index_name,
                                      scroll='10m',
                                      size=500,
                                      query={'query': query},
                                      fields=fields,
                                      runtime_mappings=runtime_mappings)
        return self._elastic_converter_factory().convert_list(generator)

    def get_aggregations(
        self,
        object_type: str,
        aggregations: dict,
        object_filters: DataSourceFilter | None = None,
        **kwargs
    ) -> dict:
        del kwargs

        real_index_name, query, fields, runtime_mappings = self._prepare_get_parameters(
            object_type=object_type,
            object_filters=object_filters,
        )
        resp = self.es.search(
            size=0,
            index=real_index_name,
            query=query,
            aggregations=aggregations,
            fields=fields,
            runtime_mappings=runtime_mappings
        )
        return resp['aggregations']

    def get_stats(
        self,
        object_type: str,
        stats_fields: list[str] = [],
        stats: list[str] = [],
        object_filters: DataSourceFilter | None = None,
        **kwargs
    ):
        del kwargs
        aggs = self.__get_aggs(
            object_type=object_type,
            stats_fields=stats_fields,
            stats=stats)
        agg_results = self.get_aggregations(
            object_type=object_type,
            aggregations=aggs,
            object_filters=object_filters
        )
        return self.__get_data_from_stats_aggregation(
            aggregation_result=agg_results,
            object_type=object_type,
            stats_fields=stats_fields,
            stats=stats
        )

    def __get_data_from_stats_aggregation(
            self,
            aggregation_result,
            object_type,
            stats_fields,
            stats
    ):
        stats_values = {}
        for stats_field in stats_fields:
            stats_values[stats_field] = {}
            for stat in stats:
                stat_value = aggregation_result[f'{stats_field}_{stat}']['value']
                python_type = self.attribute_types[object_type][stats_field]
                if python_type == 'datetime' and stat_value is not None \
                        and stat in ['min', 'max']:
                    stat_value = datetime.fromtimestamp(stat_value / 1000)
                stats_values[stats_field][stat] = stat_value
        return {'stats': stats_values}

    def get_group_stats(
        self,
        object_type: str,
        group_by: list[str],
        stats_fields: list[str] = [],
        stats: list[str] = [],
        object_filters: DataSourceFilter | None = None,
        **kwargs
    ) -> Iterable[dict[Any, int]]:
        del kwargs

        after_key = None
        while True:
            after_key, buckets = self.__get_group_stats_page(
                object_type,
                group_by,
                stats_fields=stats_fields,
                stats=stats,
                object_filters=object_filters,
                after_key=after_key)
            if len(buckets) == 0:
                break
            yield from buckets

    def __get_group_stats_page(
        self,
        object_type: str,
        group_by: list[str],
        stats_fields: list[str] = [],
        stats: list[str] = [],
        after_key: str | None = None,
        object_filters: DataSourceFilter | None = None,
    ):
        # This will return a potentially large set of results, so we need
        # to page through them
        aggregation = {
            'counts': {
                'composite': {
                    'sources': [{
                        field: {
                            'terms': {
                                'field': self._field_or_keyword(object_type, field)
                            }
                        }
                    } for field in group_by]
                },
            }
        }
        if stats_fields is not None:
            aggregation['counts']['aggregations'] = self.__get_aggs(
                object_type,
                stats_fields,
                stats
            )
        if after_key is not None:
            aggregation['counts']['composite']['after'] = after_key
        agg_page = self.get_aggregations(
            object_type,
            aggregations=aggregation,
            object_filters=object_filters)
        after_key, buckets = self.__get_data_from_group_stats_aggregation(
            agg_page,
            object_type,
            stats_fields,
            stats
        )
        return after_key, buckets

    def __get_aggs(
            self,
            object_type: str,
            stats_fields: list,
            stats: list
    ):
        ret = {}
        for stats_field in stats_fields:
            for stat in stats:
                agg = {stat: {'field': self._field_or_keyword(object_type, stats_field)}}
                if stat == 'union':
                    # This is a bespoke aggregation
                    agg = self.__get_union_aggregation(object_type, stats_field)
                elif stat == 'unique':
                    agg = self.__get_unique_count_aggregation(object_type, stats_field)
                elif self.attribute_types[object_type][stats_field] == 'str' \
                        and stat in ['min', 'max']:
                    agg = self.__get_string_aggregation(object_type, stats_field, stat)
                ret[f'{stats_field}_{stat}'] = agg
        return ret

    def __get_data_from_group_stats_aggregation(
            self,
            aggregation_result,
            object_type,
            stats_fields,
            stats
    ):
        # The after_key is sent back in one request and we use it as-is in the next request
        after_key = None
        if 'after_key' in aggregation_result['counts']:
            after_key = aggregation_result['counts']['after_key']
        buckets = aggregation_result['counts']['buckets']
        # all_stats looks like:
        # [{'key': {'first_group_by': 'value_of_first_group_by'}
        #           'second_group_by': 'value_of_second_group_by'},
        #   'stats': {'count': 123,
        #             'stats_field_stat': 345}}]
        all_stats = []
        for v in buckets:
            stats_values = {'count': v['doc_count']}
            for stats_field in stats_fields:
                stats_values[stats_field] = {}
                for stat in stats:
                    stat_value = v[f'{stats_field}_{stat}']['value']
                    python_type = self.attribute_types[object_type][stats_field]
                    if python_type == 'datetime' and stat_value is not None \
                            and stat in ['min', 'max']:
                        stat_value = datetime.fromtimestamp(stat_value / 1000)
                    stats_values[stats_field][stat] = stat_value
            all_stats.append({'key': v['key'], 'stats': stats_values})

        return after_key, all_stats

    def __get_union_aggregation(self, object_type, field):
        """
        This function is building up a union of all elements of a list in
        the aggregated field
        init_script: This is what is run at the start of each bucket
        map_script: This builds up a list, PER SHARD, of the elements in all
        records in the bucket
        combine_script: This just returns the per-shard list in our case
        reduce_script: This combines the per-shard lists into the final list
        See information on scripted metrics for Elastic for more details
        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        agg = {
            'scripted_metric': {
                'init_script': 'state.list = []',
                'map_script': f"""
                    for (element in doc['{field_or_keyword}']) {{
                        if (!state.list.contains(element)) {{
                            state.list.add(element)
                        }}
                    }}
                """,
                'combine_script': 'return state.list',
                'reduce_script': """
                    ArrayList ret = [];
                    for (a in states) {
                        for (element in a) {
                            if (!ret.contains(element)) {
                                ret.add(element)
                            }
                        }
                    }
                    return ret;
                """
            }
        }
        return agg

    def __get_string_aggregation(self, object_type, field, stat):
        """
        This function is calculating the min and max of a string

        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        comparator = '>'
        if stat == 'min':
            comparator = '<'
        agg = {
            'scripted_metric': {
                'init_script': 'state.stat = null',
                'map_script': f"""
                    for (ss in doc['{field_or_keyword}']) {{
                        if (state.stat == null) {{
                            state.stat = ss; continue
                        }}
                        if (ss.compareTo(state.stat) {comparator} 0) {{
                            state.stat = ss
                        }}
                    }}
                    """,
                'combine_script': 'return state.stat',
                'reduce_script': f"""
                    String ret = null;
                    for (a in states) {{
                        if (a == null) {{
                            continue
                        }}
                        if (ret == null) {{
                            ret = a;
                            continue
                        }}
                        if (a.compareTo(ret) {comparator} 0) {{
                            ret = a
                        }}
                    }}
                    return ret;
                """
            }
        }
        return agg

    def __get_unique_count_aggregation(self, object_type, field):
        """
        This function is calculating the unique values in the given field

        """
        field_or_keyword = self._field_or_keyword(object_type, field)
        agg = {
            'scripted_metric': {
                'params': {
                    'fieldName': field_or_keyword
                },
                'init_script': 'state.list = []',
                'map_script': """
                    if(doc[params.fieldName].size() > 0) {
                        state.list.add(doc[params.fieldName].value);
                    }
                    """,
                'combine_script': 'return state.list;',
                'reduce_script': """
                    Map uniqueValueMap = new HashMap();
                    int count = 0;
                    for(shardList in states) {
                        if(shardList != null) {
                            for(key in shardList) {
                                if(!uniqueValueMap.containsKey(key)) {
                                    count +=1;
                                    uniqueValueMap.put(key, key);
                                }
                            }
                        }
                    }
                    return count;
                """
            }
        }
        return agg

    def get_count(
        self,
        object_type: str,
        object_filters: DataSourceFilter | None = None,
        **kwargs
    ) -> int:
        del kwargs

        real_index_name, query, fields, runtime_mappings = self._prepare_get_parameters(
            object_type=object_type,
            object_filters=object_filters,
        )
        # We are not using es.count so that we can use runtime fields
        resp = self.es.search(
            index=real_index_name,
            track_total_hits=True,
            size=0,
            query=query,
            fields=fields,
            runtime_mappings=runtime_mappings
        )
        return resp['hits']['total']['value']

    @ttl_cache(ttl=3600)
    def _get_indices(self) -> dict[str, str]:
        # Get all as the actual indexes may not have the correct prefix
        results = self.es.indices.get_alias('*')
        aliased_indexes = {
            alias: index
            for index, aliases in results.items()
            for alias in aliases.get('aliases', {}).keys()
            if alias.startswith(self.index_prefix)
        }
        # Non-aliased indexes
        non_aliased_indexes = {
            index: index
            for index in results.keys()
            if index.startswith(self.index_prefix)
        }
        return aliased_indexes | non_aliased_indexes

    def _real_index_to_object_type(self, index: str) -> str:
        aliases = self._get_indices()
        alias = next((k for k, v in aliases.items() if v == index), None)
        return self.__get_object_type(alias) if alias else None

    @property
    def supported_types(self):
        indexes = self._get_indices()
        return [self.__get_object_type(index_name)
                for index_name in indexes.keys()]

    def __map_type(self, type_: str) -> str:
        if type_ in ['text', 'keyword']:
            return 'str'
        if type_ == 'long':
            return 'int'
        if type_ == 'date':
            return 'datetime'
        if type_ == 'boolean':
            return 'bool'
        return type_

    def _get_attribute_types_for_object_type(self, object_type: str) -> dict:
        index_or_alias_name = self.__get_index_or_alias(object_type)
        real_index_name = self._get_indices().get(index_or_alias_name)
        mapping = self.es.indices.get_mapping(index=index_or_alias_name)
        if 'properties' not in mapping[real_index_name]['mappings']:
            return {}
        properties = mapping[real_index_name]['mappings']['properties']
        standard_types = {
            'id' if property_name == 'uid' else property_name:
                self.__map_type(properties[property_name]['type'])
            for property_name in properties
            if 'type' in properties[property_name]
        }
        runtime_types = {
            name: self.__map_type(self.runtime_fields[object_type][name]['type'])
            for name in self.runtime_fields[object_type].keys()
        } if object_type in self.runtime_fields else {}
        return standard_types | runtime_types

    @property
    @ttl_cache(ttl=3600)
    def attribute_types(self) -> dict[str, dict[str, str]]:
        return {
            t: self._get_attribute_types_for_object_type(t)
            for t in self.supported_types
        }

    @property
    def relationship_config(self) -> dict[str, RelationshipConfig]:
        return self._relationship_cfg

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        **kwargs
    ) -> DataObject | None:
        del kwargs

        self.__validate_to_one_relation(source)

        if self.lazy_fetch and relationship_name in source._to_one_objects:
            return source._to_one_objects.get(relationship_name)

        to_one = self.relationship_config[source.type].to_one

        if relationship_name not in to_one:
            raise DataSourceError('Bad relationship name')

        new_source: DataObject = self.get_one(source.type, source.id)

        local_relation = new_source._to_one_objects.get(relationship_name)
        if local_relation is None:
            return None

        if self.lazy_fetch:
            return local_relation

        target_type = to_one[relationship_name]
        target_id = local_relation.id
        return self.get_one(target_type, target_id)

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        **kwargs
    ) -> Iterable[DataObject]:
        del kwargs

        if self.relationship_config is None:
            raise DataSourceError('There are no relationships defined')
        relationship_config = self.relationship_config[source.type]
        related_object_type = relationship_config.to_many[relationship_name]
        related_object_fk_attribute = relationship_config.foreign_keys[relationship_name]

        # Get all the related objects that point to this source object
        f = DataSourceFilter()
        f.and_ = {related_object_fk_attribute: {'eq': {'value': source.id}}}
        related_objects = self.get_list(related_object_type, object_filters=f)
        return related_objects

    def __validate_to_one_relation(self, source: DataObject) -> None:
        if self.relationship_config is None:
            raise DataSourceError('There are no relationships defined')

        if source.type not in self.relationship_config:
            raise DataSourceError('This type has no relationships')



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from typing import Iterable

from minio.datatypes import Object as MinioObject

from ..core import DataObject

if typing.TYPE_CHECKING:
    from ..core import DataSource


S3ApiResource = MinioObject
"""
The type of the response from S3 via minio
"""


class Parser(ABC):
    """
    Parses S3 API transfer resource `dict`s to `DataObject`
    instances
    """

    def parse_iterable(
        self,
        transfers: Iterable[S3ApiResource]
    ) -> Iterable[DataObject]:
        """
        Parses an `Iterable` of S3 API transfer resources
        """

        return (
            self.parse(t) for t in transfers
        )

    @abstractmethod
    def parse(self, transfer: S3ApiResource) -> DataObject:
        """
        Parses an individual S3 transfer resource to a
        `DataObject` instance
        """


class DefaultParser(Parser):
    def __init__(self, data_source_dict: dict[str, DataSource]) -> None:
        self.__dict = data_source_dict

    def parse(self, transfer: S3ApiResource) -> DataObject:
        # The name minio assigns to an S3 bucket response
        type_ = 'object'

        # Extract this object from the DataSource dict
        data_source = self.__get_data_source(type_)

        # Directly map each minio attribute to a DataSource attribute
        attributes = {
            'bucket_name': transfer.bucket_name,
            'last_modified': transfer.last_modified
        }

        # Construct and return the parsed DataSource using its factory
        return data_source.data_object_factory(
            type_,
            id_=transfer.object_name,
            attributes=attributes
        )

    def __get_data_source(self, type_: str) -> DataSource:
        return self.__dict[type_]



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Optional, Type

from cachetools.func import ttl_cache

from .attribute_metadata import (AttributeMetadata, DefaultAttributeMetadata)
from .data_object import DataObject


def data_source_attribute_metadata(
    data_source_config: DataObject,
) -> Type[AttributeMetadata]:
    """
    Takes a DataSourceConfig, and creates an AbstractMetadata
    implementation that refers to all of the child DataSourceConfigAttributes.
    """

    class DataSourceAttributeMetadata(DefaultAttributeMetadata):
        """
        AbstractMetadata that gets its info from one or more DataSources
        """
        @ttl_cache(ttl=3600)
        def __read_attributes_from_datasource(self):
            ret = {}
            # Recreate the data_source_config
            dsc = data_source_config._host.get_one(
                'data_source_config', data_source_config.id
            )
            for attribute in dsc.data_source_config_attributes:
                if attribute.object_type not in ret:
                    ret[attribute.object_type] = {}
                ret[attribute.object_type][attribute.name] = attribute
            return ret

        @ttl_cache(ttl=3600)
        def __read_cardinality_from_datasource(self, object_type):
            ret = {}
            try:
                stats = self.host.get_stats(
                    object_type,
                    stats=['cardinality'],
                    stats_fields=self.host.attribute_types[object_type].keys())
                for attribute in self.host.attribute_types[object_type].keys():
                    ret[attribute] = stats['stats'][attribute]['cardinality']
            except AttributeError:
                # Fall back to default cardinality if the datasource doesn't have stats
                pass
            return ret

        def __get_attribute(
                self,
                object_type: str,
                attribute_name: str) -> str:
            attributes = self.__read_attributes_from_datasource()
            if object_type in attributes:
                if attribute_name in attributes[object_type]:
                    return attributes[object_type][attribute_name]
            return None

        def get_display_name(
                self,
                object_type: str,
                attribute_name: str) -> str:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return super().get_display_name(object_type, attribute_name)
            return attribute.display_name

        def is_available_on_relationships(
                self,
                object_type: str,
                attribute_name: str) -> bool:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return False  # available attributes must be in DataSource
            return attribute.available_on_relationships

        def is_authoritative(
                self,
                object_type: str,
                attribute_name: str) -> bool:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return False  # authoritative attributes must be in DataSource
            return attribute.is_authoritative

        def get_cardinality(
                self,
                object_type: str,
                attribute_name: str) -> Optional[int]:
            cardinality = self.__read_cardinality_from_datasource(object_type)
            if attribute_name in cardinality:
                return cardinality[attribute_name]
            return super().get_cardinality(object_type, attribute_name)

        def get_description(
                self,
                object_type: str,
                attribute_name: str) -> Optional[str]:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return super().get_description(
                    object_type, attribute_name)
            return attribute.description

        def get_source(
                self,
                object_type: str,
                attribute_name: str) -> str:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return super().get_source(object_type, attribute_name)
            return attribute.source

        def get_acts_as(
                self,
                object_type: str,
                attribute_name: str) -> Optional[str]:
            attribute = self.__get_attribute(object_type, attribute_name)
            if attribute is None:
                return super().get_acts_as(object_type, attribute_name)
            return attribute.acts_as

    return DataSourceAttributeMetadata



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import re
from dataclasses import dataclass
from datetime import time
from typing import Iterable

from tol.core import DataObject, DataObjectToDataObjectOrUpdateConverter


class TimeStringToTimeConverter(DataObjectToDataObjectOrUpdateConverter):

    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        field_names: list[str]

    __slots__ = ['__config']
    __config: Config

    def __init__(self, data_object_factory, config: Config) -> None:
        super().__init__(data_object_factory)
        self.__config = config
        self._data_object_factory = data_object_factory

    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        """
        Converts string fields representing time in HH:MM (24-hour) format to Python time objects.
        If the string is not in HH:MM, tries to append ':00' and parse as HH:MM:SS.
        """

        for field_name in self.__config.field_names:
            value = data_object.attributes.get(field_name)
            if isinstance(value, str):
                match = re.match(r'^(\d{1,2}):(\d{2})(?::(\d{2}))?$', value)
                if match:
                    h, m = int(match.group(1)), int(match.group(2))
                    s = int(match.group(3)) if match.group(3) else 0
                    try:
                        data_object.attributes[field_name] = time(h, m, s)
                    except ValueError:
                        pass
        yield data_object



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import Dict, List

from tol.core import DataObject
from tol.core.validate import Validator

from .regex import Regex, RegexDict


class RegexByValueValidator(Validator):
    """
    Validates an incoming stream of `DataObject` instances
    according to the specified allowed values for a given
    key.
    """
    @dataclass(slots=True, frozen=True, kw_only=True)
    class Config:
        key_column: str
        regexes: Dict[str, List[Regex | RegexDict]]

    __slots__ = ['__config']
    config: Config

    def __init__(
        self,
        config: Config,
        **kwargs
    ) -> None:

        super().__init__()

        self.__config = self.__get_config(config)

    def _validate_data_object(
        self,
        obj: DataObject
    ) -> None:
        # Pull out value of the 'key_column' attribute
        key_column_value = obj.attributes.get(self.__config.key_column)
        if not key_column_value:
            return

        # Pull out relevant regex list based on this value: {[{'name': 'regex'}]}
        regex_list = self.__config.regexes.get(key_column_value)
        if not regex_list:
            return
        self.__validate_attribute(obj, regex_list)

    def __validate_attribute(
        self,
        obj: DataObject,
        regexes: List[Regex],
    ) -> None:
        for r in regexes:
            attribute_name = r.key
            value = obj.attributes.get(attribute_name)
            if not r.is_allowed(value):
                self.__add_result(obj, r)

    def __add_result(
        self,
        obj: DataObject,
        c: Regex,
    ) -> None:

        if c.is_error:
            self.add_error(
                object_id=obj.id,
                detail=c.detail,
                field=c.key
            )
        else:
            self.add_warning(
                object_id=obj.id,
                detail=c.detail,
                field=c.key,
            )

    def __get_config(
        self,
        config: Config,
    ) -> Config:

        # Ensure config is in Regex format
        # (as you can either pass in a list of Regex or a RegexDict,
        # which can be used to initialize a Regex)
        return self.Config(
            key_column=config.key_column,
            regexes={
                k: [
                    c if isinstance(c, Regex) else Regex(**c)
                    for c in v
                ]
                for k, v in config.regexes.items()
            }
        )



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import os

from .defaults import Defaults
from ..api_client import (
    ApiDataSource,
    create_api_datasource
)
from ..core import (
    core_data_object
)


def workflows(retries: int = 5, **kwargs) -> ApiDataSource:
    workflows = create_api_datasource(
        api_url=os.getenv('WORKFLOWS_URL', Defaults.WORKFLOWS_URL)
        + os.getenv('WORKFLOWS_API_PATH', Defaults.WORKFLOWS_API_PATH),
        token=os.getenv('WORKFLOWS_API_KEY'),
        data_prefix=os.getenv('WORKFLOWS_API_DATA_PATH', Defaults.WORKFLOWS_API_DATA_PATH),
        retries=retries
    )
    core_data_object(workflows)
    return workflows



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

from sqlalchemy.orm import Session as SqlaSession

from .database import Database
from .ds_session import SqlDataSourceSession
from .filter import DatabaseFilter
from .model import Model
from .relationship import SqlRelationshipConfig
from .sort import DatabaseSorter
from .sql_converter import DataObjectConverter, ModelConverter
from ..core import (
    AttributeMetadata,
    DataId,
    DataObject,
    DataSource,
    DataSourceFilter,
    DefaultAttributeMetadata,
    ReqFieldsTree,
    UserIdGetter,
)
from ..core.factory import DataObjectFactory
from ..core.operator import (
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    GroupStatter,
    Inserter,
    ListGetter,
    PageGetter,
    RelationWriteMode,
    Relational,
    ReturnMode,
    Upserter,
)
from ..core.relationship import RelationshipConfig
from ..core.requested_fields import requested_fields_to_tree

if typing.TYPE_CHECKING:
    from ..core.session import OperableSession


ConverterFactory = Callable[
    [DataObjectFactory, list[str] | None],
    ModelConverter,
]
BackConverterFactory = Callable[[], DataObjectConverter]
FilterFactory = Callable[[DataSourceFilter], DatabaseFilter]
SorterFactory = Callable[[Optional[str]], DatabaseSorter]


class SqlDataSource(
    Counter,
    Cursor,
    DataSource,
    Deleter,
    DetailGetter,
    GroupStatter,
    Inserter,
    ListGetter,
    PageGetter,
    Relational,
    Upserter,
):
    """
    A DataSource for manipulating DataObject instances as
    defined by Sqlalchemy models on a DB connection.
    """

    def __init__(
        self,
        db: Database,
        type_tablename_map: Dict[str, str],
        relationship_config: SqlRelationshipConfig,
        converter_factory: ConverterFactory,
        back_converter_factory: BackConverterFactory,
        filter_factory: FilterFactory,
        sorter_factory: SorterFactory,
        user_id_getter: Optional[UserIdGetter] = None,
        attribute_metadata: AttributeMetadata = DefaultAttributeMetadata,
        **kwargs
    ) -> None:
        self.__db = db
        self.__type_tablename_map = type_tablename_map
        self.__supported_types = list(type_tablename_map.keys())
        self.__relationship_config = relationship_config.to_dict()
        self.__converter_factory = converter_factory
        self.__back_converter_factory = back_converter_factory
        self.__filter_factory = filter_factory
        self.__sorter_factory = sorter_factory
        self.__all_attribute_types = self.__calculate_all_attribute_types()
        self.__set_user_id_getter(user_id_getter)
        self.write_batch_size = 100
        attribute_metadata.host = self

        super().__init__({}, attribute_metadata=attribute_metadata)

    def create_sqla_session(self) -> SqlaSession:
        return self.__db.session_factory()

    def get_session(self) -> OperableSession:
        return SqlDataSourceSession(self)

    @property
    def _default_write_mode(self) -> RelationWriteMode:
        return RelationWriteMode.SEPARATE

    @property
    def _default_return_mode(self) -> ReturnMode:
        return ReturnMode.POPULATED

    def __get_sqla_session(
        self, session: Optional[SqlDataSourceSession]
    ) -> SqlaSession:
        return (
            session._sqla_session if session is not None else self.create_sqla_session()
        )

    @property
    def attribute_types(self) -> Dict[str, str]:
        return self.__all_attribute_types

    @property
    def supported_types(self) -> List[str]:
        return self.__supported_types

    @property
    def relationship_config(self) -> Optional[Dict[str, RelationshipConfig]]:
        return self.__relationship_config

    def get_count(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[SqlDataSourceSession] = None,
    ) -> int:
        """
        Counts the number of results that are matched by the (optional) filter
        """

        tablename = self.__type_tablename_map[object_type]
        database_filter = self.__filter_factory(
            self._preprocess_filter(object_type, object_filters)
        )
        in_session = self.__get_sqla_session(session)
        total_count = self.__db.count(tablename, in_session, filters=database_filter)
        if session is None:
            in_session.close()
        return total_count

    @requested_fields_to_tree
    def get_by_id(
        self,
        object_type: str,
        object_ids: Iterable[DataId],
        session: Optional[SqlDataSourceSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[Optional[DataObject]]:
        in_session = self.__get_sqla_session(session)
        models = self.__get_model_list_by_ids(
            object_type,
            object_ids,
            in_session,
            requested_tree,
        )
        converter = self.__get_converter(requested_tree=requested_tree)
        return_list = list(converter.convert_iterable(models))
        if session is None:
            in_session.close()
        return return_list

    @requested_fields_to_tree
    def get_list_page(
        self,
        object_type: str,
        page_number: int,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        sort_by: Optional[str] = None,
        session: Optional[SqlDataSourceSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Tuple[Iterable[DataObject], int]:
        tablename = self.__type_tablename_map[object_type]
        database_filter = self.__filter_factory(
            self._preprocess_filter(object_type, object_filters)
        )
        sorter = self.__sorter_factory(sort_by)
        in_session = self.__get_sqla_session(session)
        total_count = self.__db.count(tablename, in_session, filters=database_filter)
        models = self.__get_list_page_models(
            tablename,
            filters=database_filter,
            page_number=page_number,
            page_size=page_size,
            sort_by=sorter,
            in_session=in_session,
            requested_tree=requested_tree,
        )
        converter = self.__get_converter(requested_tree=requested_tree)
        return_list = list(converter.convert_iterable(models))
        if session is None:
            in_session.close()
        return return_list, total_count

    @requested_fields_to_tree
    def get_list(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[SqlDataSourceSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        if self.can_use_cursor(object_type, object_filters):
            return self._get_list_by_cursor(
                object_type,
                object_filters=self._preprocess_filter(object_type, object_filters),
                session=session,
                requested_tree=requested_tree,
            )
        else:
            return self.__get_list_limit_offset(
                object_type,
                object_filters=object_filters,
                session=session,
                requested_tree=requested_tree,
            )

    def delete(
        self,
        object_type: str,
        object_ids: Iterable[str],
        session: Optional[SqlDataSourceSession] = None,
    ) -> None:
        tablename = self.__type_tablename_map[object_type]
        user_id = self.__user_id_getter()
        in_session = self.__get_sqla_session(session)
        for object_id in object_ids:
            self.__db.delete(tablename, object_id, in_session, user_id=user_id)
        if session is None:
            in_session.close()

    @requested_fields_to_tree
    def get_cursor_page(
        self,
        object_type: str,
        page_size: Optional[int] = None,
        object_filters: Optional[DataSourceFilter] = None,
        search_after: list[str] | None = None,
        session: Optional[OperableSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> tuple[Iterable[DataObject], list[str] | None]:
        fetched, _ = self.get_list_page(
            object_type,
            1,
            page_size=page_size,
            object_filters=self.update_cursor_filters(search_after, object_filters),
            sort_by='id',
            session=session,
            requested_tree=requested_tree,
        )

        return self.__format_cursor_page(fetched)

    def upsert_batch(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[SqlDataSourceSession] = None,
        merge_collections: bool | None = None,
        **kwargs,
    ) -> list[DataObject]:
        back_converter = self.__back_converter_factory()
        model_instances = back_converter.convert_iterable(objects)
        user_id = self.__user_id_getter()
        in_session = self.__get_sqla_session(session)
        returned_models = [
            self.__db.upsert(
                instance,
                in_session,
                user_id=user_id,
                merge_collections=merge_collections,
            )
            for instance in model_instances
        ]
        return_list = list(self.__get_converter().convert_iterable(returned_models))
        if session is None:
            in_session.close()
        return return_list

    def insert_batch(
        self,
        object_type: str,
        objects: Iterable[DataObject],
        session: Optional[SqlDataSourceSession] = None,
    ) -> Iterable[DataObject]:
        back_converter = self.__back_converter_factory()
        model_instances = back_converter.convert_iterable(objects)
        user_id = self.__user_id_getter()
        in_session = self.__get_sqla_session(session)
        inserted_list = [
            self.__db.insert(instance, in_session, user_id=user_id)
            for instance in model_instances
        ]
        return_list = list(self.__get_converter().convert_iterable(inserted_list))
        if session is None:
            in_session.close()
        return return_list

    def get_to_one_relation(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[SqlDataSourceSession] = None,
    ) -> Optional[DataObject]:
        tablename = self.__type_tablename_map[source.type]
        in_session = self.__get_sqla_session(session)
        model = self.__db.get_to_one_relation(
            tablename,
            source.id,
            relationship_name,
            in_session,
        )
        return_ = self.__get_converter().convert_optional(model)
        if session is None:
            in_session.close()
        return return_

    def get_to_many_relations(
        self,
        source: DataObject,
        relationship_name: str,
        session: Optional[SqlDataSourceSession] = None,
    ) -> Iterable[DataObject]:
        tablename = self.__type_tablename_map[source.type]
        in_session = self.__get_sqla_session(session)
        models = self.__db.get_to_many_relations(
            tablename, source.id, relationship_name, in_session
        )
        return_list = list(self.__get_converter().convert_iterable(models))
        if session is None:
            in_session.close()
        return return_list

    def get_group_stats(
        self,
        object_type: str,
        group_by: list[str],
        stats_fields: list[str] = [],
        stats: list[str] = ['min', 'max'],
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[OperableSession] = None,
    ) -> Iterable[dict[Any, int]]:
        tablename = self.__type_tablename_map[object_type]
        in_session = self.__get_sqla_session(session)
        filters = self.__filter_factory(
            self._preprocess_filter(object_type, object_filters)
        )

        return self.__db.get_group_stats(
            tablename,
            group_by,
            stats_fields,
            stats,
            in_session,
            filters=filters,
        )

    def __format_cursor_page(
        self,
        data_objects: Iterable[DataObject],
    ) -> tuple[Iterable[DataObject], list[str] | None]:
        return_list = list(data_objects)
        if return_list:
            return (return_list, [return_list[-1].id])
        else:
            return [], None

    def __calculate_all_attribute_types(self) -> dict[str, dict[str, str]]:
        tablename_type_map = {v: k for k, v in self.__type_tablename_map.items()}

        return {
            tablename_type_map[k]: self.__calculate_attribute_types(v)
            for k, v in self.__db.attribute_types_including_id.items()
        }

    def __calculate_attribute_types(self, types: dict[str, type]) -> dict[str, str]:
        return {k: v.__name__ for k, v in types.items()}

    def __get_converter(
        self,
        requested_tree: ReqFieldsTree | None = None,
    ) -> ModelConverter:
        return self.__converter_factory(
            self.data_object_factory,
            requested_tree,
        )

    def __get_list_limit_offset(
        self,
        object_type: str,
        object_filters: Optional[DataSourceFilter] = None,
        session: Optional[SqlDataSourceSession] = None,
        requested_tree: ReqFieldsTree | None = None,
    ) -> Iterable[DataObject]:
        page_number = 1

        while True:
            (results, _) = self.get_list_page(
                object_type,
                page_number,
                object_filters=object_filters,
                session=session,
                requested_tree=requested_tree,
            )
            results = list(results)

            yield from results

            if len(results) < self.get_page_size():
                return

            page_number += 1

    def __get_model_list_by_ids(
        self,
        object_type: str,
        object_ids: Iterable[DataId],
        session: SqlaSession,
        req_fields_tree: ReqFieldsTree | None,
    ) -> List[Optional[Model]]:
        return [
            self.__db.get_by_id(
                self.__type_tablename_map[object_type],
                id_,
                in_session=session,
                requested_tree=req_fields_tree,
            )
            for id_ in object_ids
        ]

    def __get_list_page_models(
        self,
        tablename: str,
        filters: Optional[DatabaseFilter],
        page_number: Optional[int],
        page_size: Optional[int],
        sort_by: Optional[DatabaseSorter],
        in_session: SqlaSession,
        requested_tree: ReqFieldsTree,
    ) -> Iterable[Model]:
        offset = self.__get_offset(page_number, page_size)
        return self.__db.get_page(
            tablename,
            in_session,
            filters=filters,
            sort_by=sort_by,
            offset=offset,
            limit=page_size,
            requested_tree=requested_tree,
        )

    def __get_offset(
        self, page_number: Optional[int], page_size: Optional[int]
    ) -> Optional[int]:
        return (
            None
            if page_number is None or page_size is None
            else (page_number - 1) * page_size
        )

    def __set_user_id_getter(self, user_id_getter: UserIdGetter) -> None:
        self.__user_id_getter = (
            (lambda: None) if user_id_getter is None else user_id_getter
        )



# SPDX-FileCopyrightText: 2026 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Optional


class ActionParameters:
    """
    Parses the parameters from a query string for a Action POST
    endpoint.
    """

    def __init__(self, request_args: Dict[str, str]) -> None:
        self.__request_args = request_args

    @property
    def action_name(self) -> str:  # noqa A003
        """
        The optional action name.
        """
        return self.__request_args.get('action_name')

    @property
    def ids(self) -> list[str]:  # noqa A003
        """
        The action IDs.
        """
        return self.__request_args.get('ids')

    @property
    def params(self) -> Optional[dict[str, Any]]:  # noqa A003
        """
        The optional action parameters.
        """
        if self.__request_args.get('params') is None:
            return {}
        return self.__request_args.get('params')



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

import json
import os

from ..core import (
    core_data_object
)
from ..google_sheets import (
    GoogleSheetDataSource
)


def genome_notes(**kwargs) -> GoogleSheetDataSource:
    gsds = GoogleSheetDataSource({
        'client_secrets': json.loads(os.getenv('GOOGLE_CLIENT_SECRETS')),
        'sheet_key': '1OwMsyI8a5WfQf2Y0LHM-SqJ17exjTLjjUDG-ig3aHw0',
        'mappings': {
            'genome_note': {
                'worksheet_name': 'Published record',
                'columns': {
                    'id': {
                        'heading': 'doi',
                        'type': 'str'
                    },
                    'species_name': {
                        'heading': 'species',
                        'type': 'str'
                    },
                    'taxid': {
                        'heading': 'taxid',
                        'type': 'int'
                    },
                    'tolid': {
                        'heading': 'tolid',
                        'type': 'str'
                    },
                    'passed_pr': {
                        'heading': 'Passed_PR',
                        'type': 'boolean'
                    },
                    'assembly_accession': {
                        'heading': 'Accession',
                        'type': 'str'
                    },
                    'date_published': {
                        'heading': 'Date_published',
                        'type': 'datetime',
                        'dayfirst': True
                    },
                    'pmid': {
                        'heading': 'PMID',
                        'type': 'int'
                    }

                },
                'header_row': 1,
                'data_start_row': 2
            }
        }
    })
    core_data_object(gsds)
    return gsds



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import typing
from abc import ABC, abstractmethod
from itertools import chain
from typing import Iterable

from .aggregator import Aggregator
from .counter import Counter
from .cursor import Cursor
from .deleter import Deleter
from .detail_getter import DetailGetter
from .group_statter import GroupStatter
from .inserter import Inserter
from .page_getter import PageGetter
from .relational import Relational
from .statter import Statter
from .updater import Updater
from .upserter import Upserter

if typing.TYPE_CHECKING:
    from ...core import DataSource


READ_OPERATOR_MAP: dict[str, type] = {
    'aggregate': Aggregator,
    'count': Counter,
    'cursor': Cursor,
    'detailGet': DetailGetter,
    'listGet': PageGetter,
    'relational': Relational,
    'stats': Statter,
    'groupStats': GroupStatter,
}


WRITE_OPERATOR_MAP: dict[str, type] = {
    'delete': Deleter,
    'insert': Inserter,
    'update': Updater,
    'upsert': Upserter
}


OperatorDict = dict[str, list[str]]


class OperatorConfig(ABC):
    """
    Determines which operator methods are supported on
    each of the supported types for a collection of
    `DataSource` instances.
    """

    @abstractmethod
    def to_dict(self) -> dict[str, OperatorDict]:
        """
        Returns a `dict` representation of the supported operators
        for each type of `DataObject`.
        """


class DefaultOperatorConfig(OperatorConfig):
    def __init__(self, *datasources: DataSource) -> None:
        self.__datasources = datasources
        self.__dumped_dict = self._get_dict()

    def to_dict(self) -> dict[str, OperatorDict]:
        return self.__dumped_dict

    def _get_dict(self) -> dict[str, OperatorDict]:
        pairs = self.__get_pairs()
        return dict(pairs)

    def __get_pairs(
        self
    ) -> Iterable[tuple[str, OperatorDict]]:

        operator_iterables = [
            self.__get_iterable_for_datasource(d)
            for d in self.__datasources
        ]

        return chain(*operator_iterables)

    def __get_iterable_for_datasource(
        self,
        datasource: DataSource
    ) -> Iterable[tuple[str, OperatorDict]]:

        operator_dict = self.__get_dict_for_datasource(
            datasource
        )

        return [
            (t, operator_dict)
            for t in datasource.supported_types
        ]

    def __get_dict_for_datasource(
        self,
        datasource: DataSource
    ) -> OperatorDict:

        write_operators = [
            k for k, v in WRITE_OPERATOR_MAP.items()
            if isinstance(datasource, v)
        ]
        read_operators = [
            k for k, v in READ_OPERATOR_MAP.items()
            if isinstance(datasource, v)
        ]

        return {
            'noauth': read_operators,
            'auth': write_operators
        }



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Optional, Tuple

from atlassian import Jira

import requests

from .converter import JiraIssues


class JiraClient:
    """
    Takes Jira API transfers and connects to a remote
    Jira API.
    """

    def __init__(
        self,
        jira_url: str,
        jira_api_key: str,
    ) -> None:
        self.jira = Jira(
            url=jira_url,
            token=jira_api_key
        )
        self.__jira_url = jira_url
        self.__jira_api_key = jira_api_key
        self.__type_mappings = {
            'number': 'float',
            'array': 'List[str]',
            'datetime': 'datetime'
        }

    def get_list_page(
        self,
        object_type: str,
        page: int,
        page_size: int,
        filter_string: Optional[str] = None
    ) -> Tuple[JiraIssues, int]:
        if object_type == 'issue':
            return self.__get_issues_page(
                page=page,
                page_size=page_size,
                filter_string=filter_string
            )

    def get_detail(
        self,
        object_type: str,
        issue_ids: list[str]
    ) -> JiraIssues:
        """
        Gets a single Jira issue
        """
        issues_string = '","'.join(issue_ids)
        issues, _ = self.get_list_page(
            object_type=object_type,
            filter_string=f'key in ("{issues_string}")',
            page_size=len(issue_ids),
            page=1
        )
        return issues

    def get_fields(self) -> dict:
        """
        Gets the fields of a Jira issue
        There doesn't seem to be a method for this on the JIRA class
        """
        response = requests.get(
            self.__jira_url + '/rest/api/latest/field',
            headers={
                'Authorization': f'Bearer {self.__jira_api_key}',
            }
        )
        fields = {}
        for field in response.json():
            jira_type = field['schema']['type'] if 'schema' in field else 'string'
            jira_item_type = field['schema']['items'] if 'schema' in field \
                and 'items' in field['schema'] else 'string'

            fields[field['id']] = {
                'display_name': field['name'],
                'system_name': field['name'].lower().replace('/', '').replace(' ', '_'),
                'type': self.__type_mappings[jira_type]
                if jira_type in self.__type_mappings else 'str',
                'jira_type': jira_type,
                'jira_item_type': jira_item_type,
                'clause_name': field['clauseNames'][0]
                if 'clauseNames' in field and len(field['clauseNames']) > 0 else field['name'],
                'relation': 'user' if jira_type == 'user' else None
            }
        return fields

    def get_issue_status_changelog(self, issue_id: str) -> dict:
        """
        Gets the issue status changelog of a Jira issue
        """
        print(f'Getting issue status changelog for {issue_id}')
        issue_status_changelog = self.jira.get_issue_status_changelog(
            issue_id=issue_id
        )
        return issue_status_changelog

    def __get_issues_page(
        self,
        page: int,
        page_size: int,
        filter_string: Optional[str] = None
    ) -> JiraIssues:
        """
        Gets a page of Jira issues
        """
        issues = self.jira.jql(
            jql=filter_string,
            start=page_size * (page - 1),
            limit=page_size,
            expand='changelog'
        )
        return issues['issues'], issues['total']



# SPDX-FileCopyrightText: 2023 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Union

from .aggregator import Aggregator
from .counter import Counter
from .cursor import Cursor
from .declare import get_operator_member_names  # noqa
from .deleter import Deleter
from .detail_getter import DetailGetter
from .enricher import Enricher
from .enum import OperatorMethod, RelationWriteMode, ReturnMode  # noqa
from .group_statter import GroupStatter
from .inserter import Inserter
from .list_getter import ListGetter
from .operator_config import OperatorDict  # noqa
from .page_getter import PageGetter
from .relational import Relational
from .statter import Statter
from .summariser import Summariser
from .updater import Updater
from .upserter import Upserter


ALL_OPERATORS = (
    Aggregator,
    Counter,
    Cursor,
    Deleter,
    Relational,
    DetailGetter,
    Summariser,
    Enricher,
    GroupStatter,
    Inserter,
    ListGetter,
    PageGetter,
    Statter,
    Updater,
    Upserter,
)


class AllOperatorType(
    Aggregator,
    Counter,
    Cursor,
    Deleter,
    # The below 2 are covered by Summariser
    #
    # Relational,
    # DetailGetter,
    Summariser,
    Enricher,
    GroupStatter,
    Inserter,
    ListGetter,
    PageGetter,
    Statter,
    Updater,
    Upserter,
):
    """
    A maximal, type-hint class that implements
    all classes in `ALL_OPERATORS`.
    """


# TODO - deduplicate using python 3.11
#
# Operator = Union[*ALL_OPERATORS]
Operator = Union[
    Aggregator,
    Counter,
    Cursor,
    Deleter,
    DetailGetter,
    Enricher,
    GroupStatter,
    Inserter,
    ListGetter,
    PageGetter,
    Relational,
    Statter,
    Summariser,
    Updater,
    Upserter,
]
"""
A type hint, indicating that any number of operators are implemented
"""



# SPDX-FileCopyrightText: 2024 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
import typing
from typing import Any, Iterable, Optional
from uuid import uuid4

from benchling_api_client.models.custom_entity_create import CustomEntityCreate
from benchling_api_client.models.naming_strategy import NamingStrategy
from benchling_api_client.v2.beta.models.worklist import Worklist
from benchling_api_client.v2.beta.models.worklist_create import WorklistCreate
from benchling_api_client.v2.beta.models.worklist_item import WorklistItem
from benchling_api_client.v2.beta.models.worklist_item_create import WorklistItemCreate
from benchling_api_client.v2.beta.models.worklist_type import WorklistType
from benchling_api_client.v2.stable.models.assay_result import AssayResult

from benchling_sdk.helpers.serialization_helpers import fields, none_as_unset
from benchling_sdk.models import (
    AssayResultCreate,
    AssayResultsBulkCreateRequest,
    AssayResultsCreateResponse,
    Box,
    BoxCreate,
    Container,
    ContainerBulkUpdateItem,
    ContainerContent,
    ContainerCreate,
    ContainerQuantity,
    ContainerQuantityUnits,
    ContainerUpdate,
    CustomEntity,
    CustomEntityBulkCreate,
    CustomEntityBulkUpdate,
    CustomEntityUpdate,
    Fields,
    Folder,
    FolderCreate,
    Location,
    LocationCreate,
    Measurement,
    MultipleContainersTransfer,
    Plate,
    PlateCreate,
)

from caseconverter import snakecase

from tol.core.core_converter import In

from ..core import Converter, DataDict, DataObject, DataSourceError, DataSourceUpdate

if typing.TYPE_CHECKING:
    from .benchling_datasource import BenchlingDataSource

BenchlingReturn = dict[str, dict[str, Any]]
BenchlingObject = CustomEntity | Folder | Worklist | Box | Plate | Container | \
    AssayResultsBulkCreateRequest | Location | WorklistItem | AssayResult
BenchlingObjectCreate = CustomEntityCreate | FolderCreate | WorklistCreate | WorklistItemCreate | \
    BoxCreate | PlateCreate | ContainerCreate | LocationCreate | Iterable[AssayResultCreate]
BenchlingObjectBulkCreate = CustomEntityBulkCreate | MultipleContainersTransfer | AssayResultCreate
BenchlingObjectUpdate = CustomEntityUpdate
BenchlingObjectBulkUpdate = CustomEntityBulkUpdate
BenchlingWrite = BenchlingObjectCreate | BenchlingObjectUpdate | BenchlingObjectBulkCreate | \
    BenchlingObjectBulkUpdate
"""Suitable as arguments to either `insert` or `update`"""


class BenchlingConverter(Converter[BenchlingReturn, DataObject]):

    def __init__(
            self,
            benchling_ds: BenchlingDataSource
    ) -> None:

        self.__ds = benchling_ds

        super().__init__()

    def convert_return_entities(
            self,
            input_list: list[BenchlingReturn]
    ) -> Iterable[DataObject]:
        if isinstance(input_list[0], AssayResultsCreateResponse):
            return self.__convert_assay_results(input_list)

        return (
            self.__convert_return(input_)
            for input_ in input_list
        )

    def convert(self, input_: BenchlingObject) -> DataObject:
        if isinstance(input_, Folder):
            return self.__convert_folder(input_)
        elif isinstance(input_, Worklist):
            return self.__convert_worklist(input_)
        elif isinstance(input_, Location):
            return self.__convert_location(input_)
        elif isinstance(input_, WorklistItem):
            return self.__convert_worklist_item(input_)
        elif isinstance(input_, CustomEntity):
            return self.__convert_custom_entity(input_)
        elif isinstance(input_, AssayResult):
            return self.__convert_assay_result(input_)
        if isinstance(input_, AssayResultsCreateResponse):
            return next(iter(self.__convert_assay_results(input_)), None)
        elif isinstance(input_, Box):
            return self.__convert_box_result(input_)
        elif isinstance(input_, Plate):
            return self.__convert_plate_result(input_)
        elif isinstance(input_, Container):
            return self.__convert_container_result(input_)
        raise ValueError(f'Unknown object type: {type(input_)}')

    def __convert_folder(self, input_: Folder) -> DataObject:
        return self.__ds.data_object_factory(
            'folder',
            id_=input_.id,
            attributes={'name': input_.name},
            to_one={
                'parent_folder': self.__ds.data_object_factory(
                    'folder',
                    input_.parent_folder_id,
                    stub=True
                ) if input_.parent_folder_id is not None else None
            }
        )

    def __convert_location(self, input_: Location) -> DataObject:
        object_type = snakecase(input_.schema.name)
        attributes = self.__convert_attributes(input_.fields, object_type)
        to_ones = self.__convert_relationships(input_.fields, object_type)
        native_to_ones = {}
        if input_.parent_storage_id is not None:
            native_to_ones['parent_location'] = self.__ds.data_object_factory(
                None,
                input_.parent_storage_id,
                stub=True,
                stub_types=[k for k, v in self.__ds.benchling_types.items() if v == 'location']
            )

        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes | {
                'name': input_.name,
                'barcode': input_.barcode,
            },
            to_one=to_ones | native_to_ones
        )

    def __convert_worklist(self, input_: Worklist) -> DataObject:
        return self.__ds.data_object_factory(
            'worklist',
            id_=input_.id,
            attributes={'name': input_.name, 'worklist_type': input_.type},
        )

    def __convert_worklist_item(self, input_: WorklistItem) -> DataObject:
        # Benchling gives back the object that has been added to the worklist
        # for now, we will return None
        return None

    def convert_worklist_items(self, worklist: Worklist) -> Iterable[DataObject]:
        # This is a bit of a hack as we don't know exactly the type of the objects
        # in the worklist. More work will be needed here eventually
        worklist_type = worklist.type
        if worklist_type == WorklistType.BIOENTITY:
            stub_types = list(self.__ds.schemas['custom_entity'].keys())
        elif worklist_type == WorklistType.CONTAINER:
            stub_types = list(self.__ds.schemas['container'].keys())
        worklist_items = worklist.worklist_items
        for worklist_item in worklist_items:
            to_ones = {
                'item': self.__ds.data_object_factory(
                    None,
                    id_=worklist_item.id,
                    stub=True,  # We set the type as a list - this is sorted out when unstubbed
                    stub_types=stub_types,
                )
            }

            yield self.__ds.data_object_factory(
                'worklist_item',
                id_=f'{worklist.id}_{worklist_item.id}',
                attributes={
                    'name': worklist_item.name,
                },
                to_one=to_ones | {
                    'worklist': self.__ds.data_object_factory(
                        'worklist',
                        worklist.id,
                        stub=True
                    )
                }
            )

    def __convert_custom_entity(self, input_: CustomEntity) -> DataObject:
        object_type = snakecase(input_.schema.name)
        attributes = self.__convert_attributes(input_.fields, object_type)
        to_ones = self.__convert_relationships(input_.fields, object_type)
        native_to_ones = {}
        if input_.folder_id is not None:
            native_to_ones['folder'] = self.__ds.data_object_factory(
                'folder',
                input_.folder_id,
                stub=True
            )
        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes,
            to_one=to_ones | native_to_ones
        )

    def __convert_assay_results(self, input_: AssayResultsCreateResponse) -> Iterable[DataObject]:
        assay_results = input_.assay_results

        return_objects = []
        for assay_result in assay_results:
            return_objects.append(
                self.__ds.data_object_factory(
                    'assay_result',
                    id_=assay_result
                )
            )

        return return_objects

    def __convert_assay_result(self, input_: AssayResult) -> DataObject:
        object_type = snakecase(input_.schema.name)
        attributes = self.__convert_attributes(input_.fields, object_type)
        to_ones = self.__convert_relationships(input_.fields, object_type)
        native_to_ones = {}

        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes,
            to_one=to_ones | native_to_ones
        )

    def __convert_box_result(self, input_: Box) -> DataObject:
        object_type = snakecase(input_.schema.name)
        attributes = input_.to_dict()

        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes
        )

    def __convert_plate_result(self, input_) -> DataObject:
        object_type = snakecase(input_.schema.name)
        attributes = input_.to_dict()

        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes
        )

    def __convert_container_result(self, input_: Container):
        object_type = snakecase(input_.schema.name)

        attributes = self.__convert_attributes(input_.fields, object_type)
        to_ones = self.__convert_relationships(input_.fields, object_type)
        native_to_ones = {}

        return self.__ds.data_object_factory(
            object_type,
            id_=input_.id,
            attributes=attributes | {
                'parent_storage_id': input_.parent_storage_id,
                'barcode': input_.barcode,
            },
            to_one=to_ones | native_to_ones
        )

    def convert_container_contents(
        self, input_: Iterable[ContainerContent]
    ) -> Iterable[DataObject]:
        for container_content in input_:
            entity = self.__convert_custom_entity(container_content.entity)

            attributes = {
                'batch': container_content.batch,
                'concentration': container_content.concentration.to_dict() if
                container_content.concentration else None
            }
            to_ones = {
                'entity': entity
            }

            yield self.__ds.data_object_factory(
                'container_content',
                id_=None,
                attributes=attributes,
                to_one=to_ones,
            )

    def __convert_return(self, input_: BenchlingReturn) -> DataObject:
        id_ = input_.pop('id', None)
        schema = input_.pop('schema', None)
        if schema is not None:
            object_type = snakecase(schema.get('name', None))

        attributes = self.__convert_return_attributes(input_, object_type)
        to_ones = self.__convert_return_relationships(input_, object_type)

        return self.__ds.data_object_factory(
            object_type,
            id_=id_,
            attributes=attributes,
            to_one=to_ones
        )

    def __convert_return_attributes(
            self,
            input_: BenchlingReturn,
            object_type: str
    ) -> dict[str, Any]:
        standard_attributes = {
            snakecase(k): v['value']
            for k, v
            in input_.get('fields', {}).items()
            if v['type'] not in ['dropdown', 'entity_link']
        }
        dropdown_attributes = {
            snakecase(k): self.__get_dropdown_values(
                object_type,
                snakecase(k),
                v.get('value')
            )
            for k, v
            in input_.get('fields', {}).items()
            if v['type'] == 'dropdown'
        }

        additional_attributes = {}
        if 'barcode' in input_:
            additional_attributes['barcode'] = input_.get('barcode')

        if 'parentStorageId' in input_:
            additional_attributes['parent_storage_id'] = input_.get('parentStorageId')

        return standard_attributes | dropdown_attributes | additional_attributes

    def __convert_return_relationships(
            self,
            input_: BenchlingReturn,
            object_type: str
    ) -> dict[str, Any]:
        benchling_type = self.__ds.benchling_types[object_type]
        return {
            snakecase(k): self.__ds.data_object_factory(
                self.__ds.schema_names[
                    self.__ds.schemas[benchling_type][object_type][snakecase(k)]['schema_id']
                ],
                v['value'],
                stub=True
            ) if v.get('value') != [] and v.get('value') is not None else None
            for k, v
            in input_.get('fields', {}).items()
            if v['type'] == 'entity_link'
        }

    def __convert_attributes(self, fields: Fields, object_type: str) -> dict[str, Any]:
        raw_attributes = self.__get_raw_attributes(fields)

        return self.__format_attributes(raw_attributes, object_type)

    def __get_raw_attributes(
            self,
            fields: Fields
    ) -> dict[str, dict[str, Any]]:

        fields_dict = fields.to_dict()
        additional_properties = fields_dict.pop('additional_properties', {})

        return {
            **fields_dict,
            **additional_properties
        }

    def __format_attributes(
            self,
            raw_attributes: dict[str, dict[str, Any]],
            object_type: str
    ) -> dict[str, Any]:

        standard_attributes = {
            snakecase(k): v.get('value')
            for k, v in raw_attributes.items()
            if v['type'] not in ['dropdown', 'entity_link']
        }
        dropdown_attributes = {
            snakecase(k): self.__get_dropdown_values(
                object_type,
                snakecase(k),
                v.get('value')
            )
            for k, v in raw_attributes.items()
            if v['type'] == 'dropdown'
        }
        return standard_attributes | dropdown_attributes

    def __get_dropdown_values(self, object_type: str, name: str, value: Any) -> list | str:
        if isinstance(value, list):
            return [
                self.__ds.get_attribute_value_options(object_type, name).get(v, None)
                for v in value
            ]
        return self.__ds.get_attribute_value_options(object_type, name).get(value, None)

    def __convert_relationships(self, fields: Fields, object_type: str) -> dict[str, Any]:
        raw_attributes = self.__get_raw_attributes(fields)
        return self.__format_relationships(raw_attributes, object_type)

    def __format_relationships(
            self,
            raw_attributes: dict[str, dict[str, Any]],
            object_type: str
    ) -> dict[str, Any]:
        benchling_type = self.__ds.benchling_types[object_type]
        return {
            snakecase(k): self.__ds.data_object_factory(
                self.__ds.schema_names[
                    self.__ds.schemas[benchling_type][object_type][snakecase(k)]['schema_id']
                ],
                v.get('value'),
                stub=True
            ) if v.get('value') != [] and v.get('value') is not None else None
            for k, v in raw_attributes.items()
            if v['type'] == 'entity_link'
        }


class DataObjectConverter(Converter[DataObject, BenchlingWrite]):
    # TODO: this needs splitting into two classes, one for create and one for update
    IGNORE_FIELD_NAMES = [
        'naming_strategy'
    ]

    def __init__(
            self,
            benchling_ds: BenchlingDataSource
    ) -> None:

        self.__ds = benchling_ds

        super().__init__()

    def convert_bulk(
            self,
            inputs: Iterable[Optional[In]],
            destination_type: str
    ) -> Iterable[Optional[BenchlingObjectBulkCreate]]:
        if destination_type in self.__ds.schemas['custom_entity'].keys():
            return self.__convert_insert_custom_entity_bulk(inputs, destination_type)
        if destination_type in self.__ds.schemas['assay_result'].keys():
            return self.__convert_insert_assay_result_bulk(inputs, destination_type)

        return self.convert_iterable(inputs)

    def convert_bulk_update(
            self,
            inputs: Iterable[Optional[In]],
            destination_type: str
    ) -> Iterable[Optional[BenchlingObjectUpdate]]:
        if destination_type in self.__ds.schemas['custom_entity'].keys():
            return self.__convert_update_custom_entity_bulk(inputs, destination_type)
        elif destination_type in self.__ds.schemas['container'].keys():
            return self.__convert_update_container_bulk(inputs, destination_type)

        raise DataSourceError(f'Cannot bulk update object type {destination_type}')

    def convert(self, input_: DataObject) -> BenchlingObjectCreate:
        if input_.type in self.__ds.schemas['custom_entity'].keys():
            return self.__convert_insert_custom_entity(input_)
        if input_.type in self.__ds.schemas['location'].keys():
            return self.__convert_location(input_)
        if input_.type in self.__ds.schemas['box'].keys():
            return self.__convert_box(input_)
        if input_.type in self.__ds.schemas['plate'].keys():
            return self.__convert_plate(input_)
        if input_.type in self.__ds.schemas['container'].keys():
            return self.__convert_container(input_)
        if input_.type in self.__ds.schemas['assay_result'].keys():
            return self.__convert_insert_assay_result_bulk([input_], input_.type)
        if input_.type == 'folder':
            return self.__convert_folder(input_)
        if input_.type == 'worklist':
            return self.__convert_worklist(input_)
        if input_.type == 'worklist_item':
            return self.__convert_worklist_item(input_)
        if input_.type == 'transfer':
            return self.__convert_transfer_item(input_)

        raise ValueError(f'Unknown object type: {input_.type}')

    def convert_update(
            self,
            input_: DataSourceUpdate,
            destination_type: str
    ) -> BenchlingObjectUpdate:
        # TODO: need to convert DataSourceUpdate into DataObjects
        # so we have consistency across board
        if destination_type in self.__ds.schemas['custom_entity'].keys():
            return self.__convert_update_custom_entity(input_, destination_type)
        elif destination_type in self.__ds.schemas['container'].keys():
            return self.__convert_update_container([input_], destination_type)

        raise ValueError(f'Unknown object type: {destination_type}')

    def __convert_folder(self, input_: DataObject) -> FolderCreate:
        return FolderCreate(
            name=input_.name,
            parent_folder_id=input_.parent_folder.id if input_.parent_folder is not None else None
        )

    def __convert_location(self, input_: DataObject) -> LocationCreate:
        return LocationCreate(
            name=input_.name,
            barcode=input_.barcode,
            parent_storage_id=input_.parent_storage_location
            if input_.parent_location is not None else None,
            schema_id=self.__ds.schema_ids[input_.type]
        )

    def __convert_box(self, input_: DataObject) -> BoxCreate:
        parent_storage_id = input_.attributes.get('parent_storage_id') \
            if input_.attributes.get('parent_storage_id', None) is not None \
            else input_.parent_storage_id.id

        return BoxCreate(
            barcode=input_.attributes.get('barcode'),
            parent_storage_id=parent_storage_id,
            schema_id=self.__ds.schema_ids[input_.type]
        )

    def __convert_plate(self, input_: DataObject) -> PlateCreate:
        parent_storage_id = input_.attributes.get('parent_storage_id') \
            if input_.attributes.get('parent_storage_id', None) is not None \
            else input_.parent_storage_id.id

        return PlateCreate(
            barcode=input_.attributes.get('barcode'),
            parent_storage_id=parent_storage_id,
            schema_id=self.__ds.schema_ids[input_.type]
        )

    def __convert_container(self, input_: DataObject) -> ContainerCreate:

        container_fields = self.__convert_fields(
            input_.type,
            input_.attributes,
            input_.to_one_relationships
        )

        return ContainerCreate(
            parent_storage_id=input_.attributes.get('parent_storage_id'),
            schema_id=self.__ds.schema_ids[input_.type],
            fields=container_fields,
            barcode=input_.attributes.get('barcode'),
        )

    def __convert_worklist(self, input_: DataObject) -> WorklistCreate:
        mappings = {
            'bioentity': WorklistType.BIOENTITY,
            'container': WorklistType.CONTAINER,
            'plate': WorklistType.PLATE,
            'batch': WorklistType.BATCH
        }
        return WorklistCreate(
            name=input_.name,
            type=mappings.get(input_.get_field_by_name('worklist_type'), WorklistType.BIOENTITY)
        )

    def __convert_worklist_item(self, input_: DataObject) -> WorklistItemCreate:
        return WorklistItemCreate(
            item_id=input_.item.id
        )

    def __convert_transfer_item(self, input_: DataObject) -> MultipleContainersTransfer:
        transfer_quantity = None
        if input_.attributes.get('transfer_quantity'):
            transfer_quantity = ContainerQuantity(
                units=ContainerQuantityUnits.UL,
                value=float(input_.attributes.get('transfer_quantity'))
            )

        source_concentration = None
        if input_.attributes.get('transfer_concentration'):
            source_concentration = Measurement(
                units='uL',
                value=float(input_.attributes.get('transfer_concentration'))
            )

        return MultipleContainersTransfer(
            destination_container_id=input_.attributes.get('destination_container_id'),
            source_entity_id=input_.attributes.get('source_entity_id'),
            transfer_quantity=none_as_unset(transfer_quantity),
            source_concentration=none_as_unset(source_concentration)
        )

    def __convert_insert_custom_entity_bulk(
            self,
            inputs: Iterable[DataObject],
            destination_type: str
    ) -> Iterable[CustomEntityBulkCreate]:
        return_iterable = []
        for input_ in inputs:
            naming_strategy = input_.attributes.get('naming_strategy', None)

            if not naming_strategy:
                naming_strategy = NamingStrategy.REPLACE_NAMES_FROM_PARTS

            entity_fields = self.__convert_fields(
                destination_type,
                input_.attributes,
                input_.to_one_relationships
            )

            return_iterable.append(
                CustomEntityBulkCreate(
                    name=uuid4().hex,
                    schema_id=self.__ds.schema_ids[input_.type],
                    registry_id=self.__ds.registry_id,
                    folder_id=input_.folder.id
                    if input_.folder is not None else self.__ds.folder_id,
                    fields=entity_fields,
                    naming_strategy=naming_strategy,
                    custom_fields=fields({})
                )
            )

        return return_iterable

    def __convert_update_custom_entity_bulk(
            self,
            inputs: Iterable[DataSourceUpdate],
            object_type: str
    ) -> Iterable[CustomEntityBulkUpdate]:
        return_iterable = []
        for input_ in inputs:
            update_id, update_dict = input_
            entity_fields = self.__convert_fields(
                object_type,
                {
                    k: v for k, v in update_dict.items()
                    if k in self.__ds.schemas['custom_entity'][object_type].keys()
                },
            )

            kwargs = {}
            if 'folder' in update_dict:
                kwargs['folder_id'] = \
                    update_dict['folder'].id if update_dict['folder'] is not None else None

            return_iterable.append(
                CustomEntityBulkUpdate(
                    id=update_id,
                    fields=entity_fields,
                    custom_fields=fields({}),
                    **kwargs
                )
            )

        return return_iterable

    def __convert_update_container_bulk(
            self,
            inputs: Iterable[DataSourceUpdate],
            object_type: str
    ) -> Iterable[ContainerBulkUpdateItem]:
        return_iterable = []
        for input_ in inputs:
            update_id, update_dict = input_
            container_fields = self.__convert_fields(
                object_type,
                {
                    k: v for k, v in update_dict.items()
                    if k in self.__ds.schemas['container'][object_type].keys()
                },
            )

            return_iterable.append(
                ContainerBulkUpdateItem(
                    container_id=update_id,
                    fields=container_fields,
                )
            )

        return return_iterable

    def __convert_insert_custom_entity(self, input_: DataObject) -> CustomEntityCreate:
        naming_strategy = input_.attributes.get('naming_strategy', None)

        if not naming_strategy:
            naming_strategy = NamingStrategy.REPLACE_NAMES_FROM_PARTS

        entity_fields = self.__convert_fields(
            input_.type,
            input_.attributes,
            input_.to_one_relationships
        )
        return CustomEntityCreate(
            name=uuid4().hex,
            schema_id=self.__ds.schema_ids[input_.type],
            registry_id=self.__ds.registry_id,
            folder_id=input_.folder.id if input_.folder is not None else self.__ds.folder_id,
            fields=entity_fields,
            naming_strategy=naming_strategy,
        )

    def __convert_insert_assay_result_bulk(
            self,
            inputs: Iterable[DataObject],
            object_type: str
    ) -> Iterable[AssayResultCreate]:
        return_iterable = []
        for input_ in inputs:
            assay_fields = self.__convert_fields(
                object_type,
                input_.attributes
            )

            return_iterable.append(
                AssayResultCreate(
                    schema_id=self.__ds.schema_ids[input_.type],
                    fields=assay_fields
                )
            )

        return return_iterable

    def __convert_update_custom_entity(
            self,
            input_: DataSourceUpdate,
            destination_type: str
    ) -> CustomEntityUpdate:
        update_id, update_dict = input_
        entity_fields = self.__convert_fields(
            destination_type,
            {
                k: v for k, v in update_dict.items()
                if k in self.__ds.schemas['custom_entity'][destination_type].keys()
            },
        )

        kwargs = {}
        if 'folder' in update_dict:
            kwargs['folder_id'] = \
                update_dict['folder'].id if update_dict['folder'] is not None else None

        return CustomEntityUpdate(
            fields=entity_fields,
            schema_id=self.__ds.schema_ids[destination_type],
            **kwargs
        )

    def __convert_update_container(
            self,
            input_: DataSourceUpdate,
            destination_type: str
    ) -> ContainerUpdate:
        update_id, update_dict = input_[0]
        container_fields = self.__convert_fields(
            destination_type,
            {
                k: v for k, v in update_dict.items()
                if k in self.__ds.schemas['container'][destination_type].keys()
            },
        )

        return ContainerUpdate(
            fields=container_fields,
        )

    def __convert_fields(
            self,
            object_type: str,
            data_dict: DataDict,
            to_one_relationships: dict[str, DataObject] = {}
    ) -> Fields:
        benchling_type = self.__ds.benchling_types[object_type]
        mapped_dict = {
            self.__get_field_name(object_type, name): {'value': self.__format_date(value)}
            for name, value in data_dict.items()
            if (
                name not in self.IGNORE_FIELD_NAMES
                and name in self.__ds.schemas[benchling_type][object_type].keys()
                and self.__ds.schemas[benchling_type]
                [object_type][name]['benchling_type'] != 'dropdown'
            )
        }
        mapped_relationships = {
            self.__get_field_name(object_type, name): {'value': value.id}
            for name, value in to_one_relationships.items()
            if (
                value is not None  # Not sure how to handle null relationships
                and name not in self.IGNORE_FIELD_NAMES
                and name in self.__ds.schemas[benchling_type][object_type].keys()
            )
        }
        # Convert dropdown values to Benchling dropdown values
        dropdown_values = {
            self.__get_field_name(object_type, name): {
                'value': next(
                    (k for k, v in self.__ds.get_attribute_value_options(object_type, name).items()
                     if v == value),
                    None
                )
            }
            for name, value in data_dict.items()
            if (
                name not in self.IGNORE_FIELD_NAMES
                and name in self.__ds.schemas[benchling_type][object_type].keys()
                and self.__ds.schemas[benchling_type]
                [object_type][name]['benchling_type'] == 'dropdown'
            )
        }
        return fields(mapped_dict | mapped_relationships | dropdown_values)

    def __get_field_name(
            self,
            object_type: str,
            name: str
    ) -> str:
        benchling_type = self.__ds.benchling_types[object_type]
        return self.__ds.schemas[benchling_type][object_type][name]['name']

    def __format_date(self, date: Any) -> str:
        if isinstance(date, datetime.datetime):
            return date.isoformat()
        return date



# SPDX-FileCopyrightText: 2025 Genome Research Ltd.
#
# SPDX-License-Identifier: MIT

from typing import Iterable

from ...core import (
    DataObject,
    DataObjectToDataObjectOrUpdateConverter
)


class InformaticsTolidToElasticTolidConverter(
        DataObjectToDataObjectOrUpdateConverter):
    def convert(self, data_object: DataObject) -> Iterable[DataObject]:
        if data_object.status_summary is not None:
            yield self._data_object_factory(
                'tolid',
                data_object.id,
                attributes=data_object.attributes
            )
